# -*- coding: utf-8 -*-
"""
===================================================================
Abstract representation of Atoms (:mod:`sknano.chemistry._atoms`)
===================================================================

.. currentmodule:: sknano.chemistry._atoms

.. todo::

   Apply more consistent use of upper and lower case vars for `Atoms`
   attributes.

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import copy
import math

from collections import OrderedDict, MutableSequence

import numpy as np

try:
    from scipy.spatial import KDTree
    has_kdtree = True
except ImportError:
    print('Install scipy version >= 0.13.0 to allow '
          'nearest-neighbor queries between `Atoms`.')
    has_kdtree = False

from ..tools import xyz_axes

__all__ = ['Atoms']


class Atoms(MutableSequence):

    """Class for creating abstract represention of a collection of `Atoms`.

    Parameters
    ----------
    atoms : {None, sequence, `Atoms`}, optional
        if not `None`, then a list of `Atom` instance objects or an
        existing `Atoms` instance object.
    copylist : bool, optional
        perform shallow copy of atoms list
    deepcopy : bool, optional
        perform deepcopy of atoms list
    use_kdtree : bool, optional
        use :py:class:`~scipy:scipy.spatial.KDTree` to perform
        nearest-neighbor analysis.

    """

    def __init__(self, atoms=None, copylist=True, deepcopy=False,
                 use_kdtree=True):

        self._atoms = []
        self._atom_ids = []
        self._molecule_ids = []
        self._charges = []
        self._coords = []
        self._masses = []
        self._velocities = []
        self._forces = []
        self._symbols = []
        self._coordination_numbers = []
        self._nearest_neighbors = []

        self._property_lists = \
            {'m': self._masses, 'q': self._charges, 'r': self._coords,
             'v': self._velocities, 'atomID': self._atom_ids,
             'moleculeID': self._molecule_ids, 'symbol': self._symbols,
             'CN': self._coordination_numbers, 'NN': self._nearest_neighbors}

        self._atomtypes = {}

        self._atom_tree = None

        if use_kdtree and has_kdtree is False:
            use_kdtree = False
        self._use_kdtree = use_kdtree
        self._NN_number = 6
        self._NN_cutoff = np.inf

        if atoms is not None:
            try:
                if copylist and not deepcopy:
                    self._atoms.extend(atoms.atoms[:])
                elif deepcopy:
                    self._atoms.extend(copy.deepcopy(atoms.atoms))
                else:
                    self._atoms.extend(atoms.atoms)
            except AttributeError:
                if isinstance(atoms, list):
                    if copylist and not deepcopy:
                        self._atoms = atoms[:]
                    elif deepcopy:
                        self._atoms = copy.deepcopy(atoms)
                    else:
                        self._atoms = atoms

                    for atom in self._atoms:
                        for p, plist in self._property_lists.iteritems():
                            plist.append(getattr(atom, p))
                else:
                    raise TypeError('`Atoms(atoms={!r})` '.format(atoms) +
                                    'is not a valid `Atoms` constructor '
                                    'argument.\n atoms must be `None`, a list '
                                    'of `Atom` objects, or an `Atoms` object '
                                    'instance.')

            if use_kdtree:
                self._atom_tree = self.atom_tree

    def __str__(self):
        """Return string representation of `Atoms`."""
        atoms_str = ''
        for atom in self._atoms:
            atoms_str += str(atom)
        return atoms_str

    @property
    def atoms(self):
        """Return the list of `Atom` objects"""
        return self._atoms

    @property
    def atomlist(self):
        """Return the list of `Atom` objects

        .. deprecated:: 0.1.0
           Use :py:attr:`atoms` attribute.

        """
        return self.atoms

    @property
    def atomtypes(self):
        """Return the atom type dictionary."""
        for atom in self._atoms:
            if atom.atomtype not in self._atomtypes:
                self._atomtypes[atom.atomtype] = {}
                self._atomtypes[atom.atomtype]['mass'] = atom.m
                self._atomtypes[atom.atomtype]['q'] = atom.q
        return self._atomtypes

    @property
    def atom_ids(self):
        """Return array of `Atom` IDs."""
        atom_ids = []
        for atom in self._atoms:
            atom_ids.append(atom.atomID)
        self._atom_ids = atom_ids[:]
        return np.asarray(self._atom_ids)

    @property
    def atom_tree(self):
        """Return the :py:class:`~scipy:scipy.spatial.KDTree` of coords."""
        if self._use_kdtree and len(self._atoms) != 0:
            self._atom_tree = KDTree(self.coords)
        return self._atom_tree

    @property
    def CM(self):
        """Center-of-Mass coordinates of `Atoms`.

        Returns
        -------
        ndarray
            3-element ndarray specifying center-of-mass coordinates of `Atoms`.

        """
        masses = np.asarray([self.masses])
        coords = np.asarray(self.coords)
        MxR = masses.T * coords
        return np.sum(MxR, axis=0) / np.sum(masses)

    @property
    def M(self):
        """Total mass of `Atoms`."""
        return math.fsum(self.masses)

    @property
    def Natoms(self):
        """Number of atoms in `Atoms`."""
        return len(self._atoms)

    @property
    def NN_cutoff(self):
        return self._NN_cutoff

    @NN_cutoff.setter
    def NN_cutoff(self, value):
        self._NN_cutoff = value

    @property
    def NN_number(self):
        return self._NN_number

    @NN_number.setter
    def NN_number(self, value):
        self._NN_number = int(value)

    @property
    def Ntypes(self):
        """Number of atom types in `Atoms`."""
        return len(self.atomtypes.keys())

    @property
    def charges(self):
        """Return list of `Atom` charges."""
        #self._charges_array = np.asarray(self._charges)
        charges = []
        for atom in self._atoms:
            charges.append(atom.q)
        self._charges = charges[:]
        return np.asarray(self._charges)

    @property
    def coords(self):
        """Return array of `Atom` coordinates."""
        coords = []
        for atom in self._atoms:
            coords.append(atom.r)
        self._coords = coords[:]
        return np.asarray(self._coords)

    @property
    def coordination_numbers(self):
        """Return array of `Atom` coordination numbers."""
        self.update_coordination_numbers()
        coordination_numbers = []
        for atom in self._atoms:
            coordination_numbers.append(atom.CN)
        self._coordination_numbers = coordination_numbers[:]
        return np.asarray(self._coordination_numbers)

    def update_coordination_numbers(self):
        """Update `Atom` coordination numbers."""
        if self._use_kdtree:
            NN_d, NN_i = \
                self.query_atom_tree(n=self.NN_number,
                                     cutoff_radius=self.NN_cutoff)
            for i, atom in enumerate(self._atoms):
                CN = 0
                for d in NN_d[i]:
                    if d < self.NN_cutoff:
                        CN += 1
                atom.CN = CN

    @property
    def nearest_neighbors(self):
        """Return array of nearest-neighbor atoms for each `Atom`."""
        self.update_nearest_neighbors()
        nearest_neighbors = []
        for atom in self._atoms:
            nearest_neighbors.append(atom.NN)
        self._nearest_neighbors = nearest_neighbors[:]
        return np.asarray(self._nearest_neighbors)

    def update_nearest_neighbors(self):
        """Update `Atom` nearest-neighbors."""
        if self._use_kdtree:
            NN_d, NN_i = self.query_atom_tree(n=self.NN_number,
                                              cutoff_radius=self.NN_cutoff)
            for i, atom in enumerate(self._atoms):
                NN_atoms = []
                for j, d in enumerate(NN_d[i]):
                    if d < self.NN_cutoff:
                        NN_atoms.append(self._atoms[NN_i[i][j]])
                atom.NN = NN_atoms

    def query_atom_tree(self, n=6, eps=0, p=2, cutoff_radius=np.inf):
        """Query atom tree for nearest neighbors distances and indices.

        Parameters
        ----------
        n : integer
            The number of nearest neighbors to return.
        eps : nonnegative float
            Return approximate nearest neighbors; the kth returned value
            is guaranteed to be no further than (1+eps) times the
            distance to the real kth nearest neighbor.
        p : float, 1<=p<=infinity
            Which Minkowski p-norm to use.
            1 is the sum-of-absolute-values "Manhattan" distance
            2 is the usual Euclidean distance
            infinity is the maximum-coordinate-difference distance
        cutoff_radius : nonnegative float
            Return only neighbors within this distance. This is used to prune
            tree searches, so if you are doing a series of nearest-neighbor
            queries, it may help to supply the distance to the nearest neighbor
            of the most recent point.

        Returns
        -------
        NN_d : array of floats
            The distances to the nearest neighbors.
        NN_i : array of integers
            The locations of the neighbors in self.atom_tree.data. NN_i is the
            same shape as NN_d.

        """
        atom_tree = self.atom_tree
        NN_d = NN_i = None
        if atom_tree is not None:
            d, i = atom_tree.query(self.coords, k=n+1, eps=eps, p=p,
                                   distance_upper_bound=cutoff_radius)
            NN_d, NN_i = d[:, 1:], i[:, 1:]
        return NN_d, NN_i

    def query_coordination_numbers(self, n=6, rc=np.inf):
        """Query and update `Atom` coordination numbers.

        Parameters
        ----------
        n : int, optional
        rc : nonnegative float, optional

        """
        if self._use_kdtree:
            NN_d, NN_i = self.query_atom_tree(n=n, cutoff_radius=rc)
            for i, atom in enumerate(self._atoms):
                CN = 0
                for d in NN_d[i]:
                    if d < rc:
                        CN += 1
                atom.CN = CN
        coordination_numbers = []
        for atom in self._atoms:
            coordination_numbers.append(atom.CN)
        self._coordination_numbers = coordination_numbers[:]
        return np.asarray(self._coordination_numbers)

    def query_nearest_neighbors(self, n=6, rc=np.inf):
        """Query and update `Atom` nearest neighbors.

        Parameters
        ----------
        n : int, optional
        rc : nonnegative float, optional

        """
        if self._use_kdtree:
            NN_d, NN_i = self.query_atom_tree(n=n, cutoff_radius=rc)
            for i, atom in enumerate(self._atoms):
                NN_atoms = []
                for j, d in enumerate(NN_d[i]):
                    if d < rc:
                        NN_atoms.append(self._atoms[NN_i[i][j]])
                atom.NN = NN_atoms
        nearest_neighbors = []
        for atom in self._atoms:
            nearest_neighbors.append(atom.NN)
        self._nearest_neighbors = nearest_neighbors[:]
        return np.asarray(self._nearest_neighbors)

    @property
    def masses(self):
        """Return the list of `Atom` masses."""
        #self._masses_array = np.asarray(self._masses)
        masses = []
        for atom in self._atoms:
            masses.append(atom.m)
        self._masses = masses[:]
        return self._masses

    @property
    def m(self):
        """Total mass of `Atoms`.

        .. deprecated:: 0.1.0
           Use :py:attr:`M` attribute instead.

        """
        return self.M

    @property
    def q(self):
        """Return the total net charge of `Atoms`."""
        return np.asarray(self.charges).sum()

    @property
    def r(self):
        """Alias for :py:attr:`~sknano.chemistry.Atoms.CM`"""
        return self.CM

    @property
    def symbols(self):
        """Return array of `Atom` symbols."""
        symbols = []
        for atom in self._atoms:
            symbols.append(atom.symbol)
        self._symbols = symbols[:]
        return np.asarray(self._symbols)

    @property
    def velocities(self):
        """Return array of `Atom` velocities."""
        velocities = []
        for atom in self._atoms:
            velocities.append(atom.v)
        self._velocities = velocities[:]
        return np.asarray(self._velocities)

    def add_atomtype(self, atom):
        """Add atom type to :py:attr:`~sknano.chemistry.Atoms.atomtypes` dict.

        Parameters
        ----------
        atom : `Atom`
            an intance of an `Atom` object

        """
        if atom.atomtype not in self._atomtypes:
            self._atomtypes[atom.atomtype] = {}
            self._atomtypes[atom.atomtype]['mass'] = atom.m
            self._atomtypes[atom.atomtype]['q'] = atom.q

    def add_atomtypes(self, atomtypes=[]):
        """Add atom type in atomtypes list to atomtype dictionary.
        Parameters
        ----------
        atomtypes : sequence
            a list of `Atom` object instances

        """
        for atom in atomtypes:
            self.add_atomtype(atom)

    def assign_unique_ids(self, starting_id=1):
        """Assign unique ID to each `Atom` in `Atoms`."""
        for i, atom in enumerate(self._atoms, start=starting_id):
            atom.atomID = i

    def center_CM(self, r_indices=[0, 1, 2]):
        """Center atoms on CM coordinates

        Parameters
        ----------
        r_indices : sequence, optional
            list of component indices of CM position vector to translate
            and center on the origin.

        """
        dr = -self.CM
        self.translate(dr, r_indices=r_indices)

    def clip_bounds(self, min_limit=None, max_limit=None,
                    abs_limit=None, r_indices=[0, 1, 2]):
        """Remove atoms outside the given limits along given dimension.

        Parameters
        ----------
        min_limit : {None, float}, optional
        max_limit : {None, float}, optional
        abs_limit : {None, float}, optional
        r_indices : sequence, optional

        """
        for atom in self._atoms[:]:
            r = atom.r.tolist()
            for i, ri in enumerate(r[:]):
                if i in r_indices:
                    if abs_limit is not None:
                        if abs(ri) <= abs_limit:
                            continue
                        else:
                            del self._atoms[self._atoms.index(atom)]
                            break
                    else:
                        if min_limit is not None:
                            if ri >= min_limit:
                                continue
                            else:
                                del self._atoms[self._atoms.index(atom)]
                                break
                        if max_limit is not None:
                            if ri <= max_limit:
                                continue
                            else:
                                del self._atoms[self._atoms.index(atom)]
                                break

    def filter_atoms(self, filter_atom_ids, invert=False):
        """Filter `Atoms`.

        Parameters
        ----------
        filter_atom_ids : array_like
        invert : bool, optional

        Returns
        -------
        filtered_atoms : `Atoms`

        """
        #if invert:
        #    filtered_atoms = \
        #        Atoms([atom for atom in self.atoms if
        #               atom.atomID not in filter_atom_ids])

        #else:
        #    filtered_atoms = \
        #        Atoms([atom for atom in self.atoms if
        #               atom.atomID in filter_atom_ids])

        filter_indices = \
            np.in1d(self.atom_ids, filter_atom_ids, invert=invert).nonzero()
        filtered_atoms = \
            Atoms(np.asarray(self._atoms)[filter_indices].tolist())
        return filtered_atoms

    def fix_minus_zero_coords(self, epsilon=1.0e-10):
        """Set really really small negative coordinates to zero.

        Set all coordinates with absolute value less than
        epsilon zero so we don't end up with -0.00000
        coordinates in structure data output.

        Parameters
        ----------
        epsilon : float
            smallest allowed absolute value of any component of an
            atom's position

        """
        for atom in self._atoms:
            r = atom.r.tolist()
            for i, ri in enumerate(r[:]):
                if ri < 0 and abs(ri) < epsilon:
                    r[i] = 0.0
            atom.x, atom.y, atom.z = r

    def getatomsattr(self, asarray=False, as_array=False):
        pass

    def get_atom(self, atomID=None, index=None):
        try:
            return self._atoms[atomID - 1]
        except (TypeError, IndexError):
            try:
                return self._atoms[index]
            except (TypeError, IndexError):
                return None

    def get_atoms(self, asarray=False, as_array=False):
        """Return list of `Atoms`.

        Parameters
        ----------
        asarray, as_array : bool, optional

        Returns
        -------
        sequence or ndarray

        """
        if asarray or as_array:
            return np.asarray(self._atoms)
        else:
            return self._atoms

    def get_coords(self, components=None, as_dict=False):
        """Return atom coords.

        Parameters
        ----------
        components : {None, sequence}, optional
        as_dict : bool, optional

        Returns
        -------
        coords : :py:class:`python:~collections.OrderedDict` or ndarray

        """
        coords = self.coords
        if as_dict:
            if components is None or components == 'r':
                components = ('x', 'y', 'z')
            elif isinstance(components, str):
                components = (components,)

            return OrderedDict(zip(
                components, [coords[:, xyz_axes.index(component)] for
                             component in components]))
        else:
            return coords

    def get_filtered_atom_ids(self, filter_atom_ids, invert=False):
        """Return atom ids filtered by list of `filter_atom_ids`.

        Parameters
        ----------
        invert : bool, optional

        Returns
        -------
        ndarray

        """
        atom_ids = self.atom_ids
        filter_indices = \
            np.in1d(atom_ids, filter_atom_ids, invert=invert).nonzero()
        return atom_ids[filter_indices]

    def get_filtered_coords(self, filter_atom_ids, components=None,
                            as_dict=False, invert=False):
        """Return filtered coordinates filtered by filter_atom_ids.

        Parameters
        ----------
        filter_atom_ids : array_like
        components : {None, sequence}, optional
        as_dict : bool, optional
        invert : bool, optional

        Returns
        -------
        filtered_coords : :py:class:`python:~collections.OrderedDict` or
        ndarray

        """
        atom_ids = self.atom_ids
        coords = self.coords
        filter_indices = \
            np.in1d(atom_ids, filter_atom_ids, invert=invert).nonzero()
        filtered_coords = coords[filter_indices]

        if components is None or components == 'r':
            components = ('x', 'y', 'z')
        elif isinstance(components, str):
            components = (components,)

        if as_dict:
            return OrderedDict(zip(
                components, [filtered_coords[:, xyz_axes.index(component)]
                             for component in components]))
        else:
            filtered_coords

    def rezero_coords(self, epsilon=1.0e-10):
        """Set really really small coordinates to zero.

        Set all coordinates with absolute value less than
        epsilon to zero.

        Parameters
        ----------
        epsilon : float
            smallest allowed absolute value of any :math:`x,y,z` component.

        """
        for atom in self._atoms:
            r = atom.r.tolist()
            for i, ri in enumerate(r[:]):
                if abs(ri) < epsilon:
                    r[i] = 0.0
            atom.x, atom.y, atom.z = r

    def rezero_xyz(self, epsilon=1.0e-10):
        return self.rezero_coords(epsilon=epsilon)

    def rotate(self, R_matrix):
        """Rotate atom coordinates using rotation matrix `R_matrix`.

        Parameters
        ----------
        R_matrix : array_like
            3x3 array representation of *3D rotation matrix*.

        """
        for atom in self._atoms:
            atom.r = np.dot(R_matrix, atom.r.T).T

    def translate(self, dr, r_indices=[0, 1, 2]):
        """Translate atom coordinates.

        Parameters
        ----------
        dr : array_like
            array representation of displacement vector to translate
            components of atoms position vector through
        r_indices : sequence, optional
            list of component indices of position vector that
            `dr` vector acts upon

        """
        for atom in self._atoms:
            r = atom.r.tolist()
            for i, ri in enumerate(r[:]):
                if i in r_indices:
                    r[i] += dr[i]
            atom.x, atom.y, atom.z = r

    def __delitem__(self, index):
        """Concrete implementation of @abstractmethod.

        Delete list element `self.atoms[index]` and delete all elements
        from atom properties lists `self.masses[index]`,
        `self.charges[index]`, and `self.coords[index]`

        Parameters
        ----------
        index : int
            index of target list element

        """
        del self._atoms[index]
        for plist in self._property_lists.itervalues():
            del plist[index]

    def __getitem__(self, index):
        """Concrete implementation of @abstractmethod.

        Get `Atom` object instance at list element `self.atoms[index]`

        Parameters
        ----------
        index : int
            index of target list element

        Returns
        -------
        `Atom`
            `Atom` instance at target `self.atoms[index]`

        """
        return self._atoms[index]

    def __len__(self):
        """Concrete implementation of @abstractmethod.

        Returns
        -------
        int
            length of `self.atoms` list.

        """
        return len(self._atoms)

    def __setitem__(self, index, atom):
        """Concrete implementation of @abstractmethod.

        set target list element `self.atoms[index] = atom`

        Also set element of all atom properties lists (`self.masses[index]`,
        `self.charges[index]`, and `self.coords[index]`) to atom instance
        properties (`atom.m`, `atom.q`, `atom.r`), respectively.

        Parameters
        ----------
        index : int
            index of target list element
        atom : `Atom`
            `Atom` instance object to set target list element to.

        """
        self._atoms[index] = atom
        for p, plist in self._property_lists.iteritems():
            plist[index] = getattr(atom, p)

    def insert(self, index, atom):
        """Concrete implementation of @abstractmethod.

        insert `Atom` instance at target list `index`

        Also insert `Atom` instance properties at the given target list index
        for all `Atom` properties in `self._property_lists.keys()`
        into their respective target lists of `Atoms` properties
        `self._property_lists.values()`.

        Parameters
        ----------
        index : int
            index of target list element
        atom : `Atom`
            `Atom` object instance to set target list element to

        """
        self._atoms.insert(index, atom)
        for p, plist in self._property_lists.iteritems():
            plist.insert(index, getattr(atom, p))



# -*- coding: utf-8 -*-
"""
==============================================================
Preset data structures (:mod:`sknano.tools._luts`)
==============================================================

.. currentmodule:: sknano.tools._luts

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

__all__ = ['chiral_type_name_mappings',
           'components', 'dimensions',
           'xyz', 'xyz_axes']

chiral_type_name_mappings = \
    {'achiral': 'aCh', 'armchair': 'AC', 'zigzag': 'ZZ', 'chiral': 'Ch'}

components = dimensions = xyz = xyz_axes = ('x', 'y', 'z')



# -*- coding: utf-8 -*-
"""
=========================================================
TubeGen wrapper class (:mod:`sknano.nanogen._tubegen`)
=========================================================

.. currentmodule:: sknano.nanogen._tubegen

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import os
import subprocess
import sys

#from collections import OrderedDict

import numpy as np

from ..structure_io import XYZReader, XYZWriter
from ..tools import plural_word_check
from ..tools.refdata import CCbond

from ._nanotubes import Nanotube

format_ext = {'gaussian': '.com',
              'gaussian-pbc': '.com',
              'wien': '.struct',
              'xyz': '.xyz',
              'pdb': '.pdb',
              'pdb-pbc': '.pdb',
              'pov': '.pov',
              'bgf': '.bgf',
              'cif': '.cif'}

shape_structure_dict = {'nanotube': 'hexagonal',
                        'graphene': 'planar',
                        'hexagonal-bundle': 'hexagonal',
                        'cubic-bundle': 'cubic',
                        'custom': None}

__all__ = ['TubeGen', 'TubeGenArgs', 'format_ext', 'shape_structure_dict']


class TubeGenArgs(object):
    """A class for populating tubegen arg attributes"""
    pass


class TubeGen(object):

    """Wrapper class around the TubeGen binary.

    TubeGen ([TG]_) is developed by Jeffrey T. Frey.

    Parameters
    ----------
    fmt : str, optional
        Structure data format. Must be one of:

            - xyz
            - gaussian
            - gaussian-pbc
            - wien
            - pdb
            - pdb-pbc
            - pov
            - bgf
            - cif

    units : {'angstrom', 'bohr'}, optional
        The unit of length to generate the spatial coordinates in.
    bond : float, optional
        Bond length between `element1` and `element2`.
    element1, element2 : {int, str}, optional
        Atomic {number, symbol} of two atom basis.
    gutter : sequence, optional
        A 3-tuple of floats specifying the unit cell gutter in **Angstroms**.
    shape : {'hexagonal', 'cubic', 'planar'}, optional
        Selects the structure shape.
    chirality : sequence, optional
        A 2-tuple of ints specifying the chiral indices defining the
        chirality :math:`\\mathbf{C}_{h} = (n, m)`
    cell_count : sequence, optional
        A 3-tuple of ints specifying the unit cell count along the
        :math:`(x, y, z)` axes.
    relax_tube : {'yes', 'no'}, optional
        Selects whether or not to apply energy minimization of the structure.
    tube_length : float, optional
        length of the nanotube in **nanometers**.

    Examples
    --------

    >>> from sknano.nanogen import TubeGen

    Generate a single unit cell of a :math:`\\mathbf{C}_{h} = (10, 5)`
    :abbr:`SWCNT (single-walled carbon nanotube)` in `xyz` format:

    .. code-block:: python

       >>> swcnt1005 = TubeGen(chirality=(10, 5))
       >>> swcnt1005.generate()
       TubeGen - Carbon Nanotube Stucture Generator
       Version 3.4 [03:16:28 Oct 29 2013]
       Copyright (c) 2000-2013, J. T. Frey

       > set element1 C
       > set element2 C
       > set gutter 1.6735,1.6735,0
       > set format xyz
       > set units angstrom
       > set cell_count 1,1,1
       > set chirality 10,5
       --- Construction of Graphitic Atomic-Basis Vectors ---------------------
       Nearest neighbor bonding distance as:  1.421
           a1 = < 2.1315 , 1.23062 >
           a2 = < 2.1315 , -1.23062 >
       ------------------------------------------------------------------------

       --- Construction of Chiral/Tubule Translation Vectors ------------------
       n = ( 10) and m = (  5):
       n'= (  4) and m'= (  5):
       Chiral vector Ch constructed as 10(a1) + 5(a2):
           Ch = < 31.9725 , 6.15311 >, |Ch| = 32.5592
       Chiral angle is 19.1066 degrees
       Tubule translation vector T constructed as 4(a1) - 5(a2):
           T = < -2.1315 , 11.0756 >, |T| = 11.2788
       Tubule radius: 5.18196
       Tubule height: 11.2788
       Reciprocal-space vectors:
           T_R = < -0.0167555 , 0.087064 >
           Ch_R = < 0.0301598 , 0.00580427 >
       ------------------------------------------------------------------------

       --- Relaxing tubule to appropriate bond lengths ------------------------
                                                   delta-radius           1e-15
                                                   scaling-factors        1e-15
                                                   error-function         1e-15
       ========================================================================
                                                                     Iterations
       Iter     delta-h      delta-r      Gammas                     a1  a2  a3
       ========================================================================
       0                               1.000000  1.000000  1.000000  45  39  30
       1     3.5044e-03   1.0648e-02   1.002936  1.000576  1.000036  26  21   8
       2    -1.5682e-05  -4.3187e-05   1.002925  1.000573  1.000036   6   0   0
       3     9.0760e-09   1.1291e-07   1.002925  1.000573  1.000036   0   0   0
       4     0.0000e+00   0.0000e+00   1.002925  1.000573  1.000036
       ========================================================================
       Convergence reached in 4 cycles
       New graphitic basis:
           a1 = < 2.13606 , 1.23133 >
           a2 = < 2.13568 , -1.23067 >
           cc-bond = 1.42516
       New chiral/tubule translation vectors:
           Ch = < 32.039 , 6.15994 >, |Ch| = 32.6258
           T = < -2.13415 , 11.0786 >, |T| = 11.2823
       Tubule radius: 5.19256     [total delta-r of 0.0106047]
       Tubule height: 11.2823     [total delta-h of 0.00348869]
       Angle between Ch and T:  97.6112 degrees
       ------------------------------------------------------------------------

       > set shape hexagonal
       > set relax_tube yes
       > set bond 1.421
       --- Construction of Graphitic Atomic-Basis Vectors ---------------------
       Nearest neighbor bonding distance as:  1.421
           a1 = < 2.1315 , 1.23062 >
           a2 = < 2.1315 , -1.23062 >
       ------------------------------------------------------------------------

       --- Construction of Chiral/Tubule Translation Vectors ------------------
       n = ( 10) and m = (  5):
       n'= (  4) and m'= (  5):
       Chiral vector Ch constructed as 10(a1) + 5(a2):
           Ch = < 31.9725 , 6.15311 >, |Ch| = 32.5592
       Chiral angle is 19.1066 degrees
       Tubule translation vector T constructed as 4(a1) - 5(a2):
           T = < -2.1315 , 11.0756 >, |T| = 11.2788
       Tubule radius: 5.18196
       Tubule height: 11.2788
       Reciprocal-space vectors:
           T_R = < -0.0167555 , 0.087064 >
           Ch_R = < 0.0301598 , 0.00580427 >
       ------------------------------------------------------------------------

       --- Relaxing tubule to appropriate bond lengths ------------------------
                                                   delta-radius           1e-15
                                                   scaling-factors        1e-15
                                                   error-function         1e-15
       ========================================================================
                                                                     Iterations
       Iter    delta-h      delta-r      Gammas                      a1  a2  a3
       ========================================================================
       0                               1.000000  1.000000  1.000000  45  39  30
       1     3.5044e-03   1.0648e-02   1.002936  1.000576  1.000036  26  21   8
       2    -1.5682e-05  -4.3187e-05   1.002925  1.000573  1.000036   6   0   0
       3     9.0760e-09   1.1291e-07   1.002925  1.000573  1.000036   0   0   0
       4     0.0000e+00   0.0000e+00   1.002925  1.000573  1.000036
       ========================================================================
       Convergence reached in 4 cycles
       New graphitic basis:
           a1 = < 2.13606 , 1.23133 >
           a2 = < 2.13568 , -1.23067 >
           cc-bond = 1.42516
       New chiral/tubule translation vectors:
           Ch = < 32.039 , 6.15994 >, |Ch| = 32.6258
           T = < -2.13415 , 11.0786 >, |T| = 11.2823
       Tubule radius: 5.19256     [total delta-r of 0.0106047]
       Tubule height: 11.2823     [total delta-h of 0.00348869]
       Angle between Ch and T:  97.6112 degrees
       ------------------------------------------------------------------------

       > generate
       Producing rolled, hexagonal nanotube lattice.

       Lattice consists of 70 hexagonal sub-cells.
       Cell generation complete.  140 basis points defined.
       > save 1005r_1cell.xyz
       > tubegen finished with no errors


    References
    ----------
    .. [TG] TubeGen 3.4 (web-interface,
            http://turin.nss.udel.edu/research/tubegenonline.html),
            J. T. Frey and D. J. Doren, University of Delaware,
            Newark DE, 2011

    """

    def __init__(self, fmt='xyz', units='angstrom', bond=CCbond,
                 element1='C', element2='C', gutter=(1.6735, 1.6735, 0),
                 shape='hexagonal', chirality=(10, 10), cell_count=(1, 1, 1),
                 relax_tube='yes', Lz=None, tube_length=None):

        n = int(chirality[0])
        m = int(chirality[1])

        T = Nanotube.compute_T(n=n, m=m, bond=bond)

        if Lz is None and tube_length is not None:
            Lz = tube_length

        nx, ny, nz = cell_count

        if Lz is not None and Lz != 'None' and \
                isinstance(Lz, (int, float, str)) and \
                shape in ('hexagonal', 'cubic'):
            Lz = 10 * float(Lz)
            nz = int(np.ceil(Lz / T))
        elif nz is not None and nz != 'None' and \
                isinstance(nz, (int, float, str)):
            nz = int(nz)
            Lz = nz * T

        self._Lz = Lz
        self._dt = Nanotube.compute_dt(n=n, m=m, bond=bond)

        self._fmt = fmt
        self._units = units
        self._bond = bond
        self._e1 = element1
        self._e2 = element2
        self._gutter = ','.join([str(x) for x in gutter])
        self._shape = shape
        self._chirality = ','.join([str(x) for x in chirality])
        self._relax_tube = relax_tube
        self._cell_count = ','.join([str(x) for x in (nx, ny, nz)])
        self._kwargs = {'format': self._fmt,
                        'units': self._units,
                        'bond': self._bond,
                        'element1': self._e1,
                        'element2': self._e2,
                        'gutter': self._gutter,
                        'shape': self._shape,
                        'chirality': self._chirality,
                        'cell_count': self._cell_count,
                        'relax_tube': self._relax_tube}
        self._genfile_name = ''
        self._output = ''

    @property
    def fmt(self):
        """Structure data format.

        Returns
        -------
        fmt : str
            Structure data format.

        """
        return self._fmt

    @property
    def cell_count(self):
        """Cell count in :math:`x, y, z` dimensions.

        Returns
        -------
        tuple
            3-tuple of integers: :math:`nx, ny, nz`

        """
        return tuple([int(c) for c in self._cell_count.split(',')])

    @property
    def chirality(self):
        """Chirality :math:`\\mathbf{C_h} = (n, m)`

        Returns
        -------
        tuple
            2-tuple of integers :math:`(n, m)`

        """
        return tuple([int(c) for c in self._chirality.split(',')])

    @property
    def dt(self):
        """Tube diameter.

        Returns
        -------
        float
            nanotube diameter in Angstroms

        """
        return self._dt

    @property
    def gutter(self):
        """Gutter in :math:`x, y, z` dimensions.

        Returns
        -------
        tuple
            3-tuple of floats

        """
        return tuple([float(c) for c in self._gutter.split(',')])

    @property
    def output(self):
        """Output file.

        Returns
        -------
        str
            output file name

        """
        return self._output

    @property
    def shape(self):
        """Structure shape.

        Returns
        -------
        str
            Structure shape will be one of {'hexagonal', 'cubic', 'planar'}.

        """
        return self._shape

    def fix_length(self):
        """Crop nanotube ends for correct length."""
        if self._Lz is not None:
            if self._fmt == 'xyz':
                xyzreader = XYZReader(self._output)
                atoms = xyzreader.atoms
                atoms.clip_bounds(abs_limit=self._Lz / 2 + 0.1, r_indices=[2])
                XYZWriter.write(fname=self._output, atoms=atoms,
                                comment_line=xyzreader.comment_line)

    def cleanup(self):
        """Cleanup tubegen input file."""
        if os.path.isfile(self._genfile_name):
            try:
                os.remove(self._genfile_name)
            except OSError as e:
                print(e)
            else:
                self._genfile_name = ''

    def generate(self):
        """Generate structure."""
        self._genfile_name = self._name_genfile()
        try:
            self._write_genfile()
        except OSError as e:
            print(e)
            sys.exit(1)

        self._tubegen()
        self._output = \
            self._genfile_name.split('.')[0] + format_ext[self._fmt]

    def _name_genfile(self):
        genfile = None
        cell_count = self._cell_count
        chirality = self._chirality
        shape = self._shape
        xcells, ycells, zcells = tuple([int(c) for c in cell_count.split(',')])
        if cell_count == '1,1,1':
            genfile = ''.join([x.zfill(2) for x in chirality.split(',')]) \
                + ('f' if shape == 'planar' else 'r') + '_1cell' + '.gen'
        elif shape != 'planar' \
                and xcells == 1 and ycells == 1 and zcells != 1:
            genfile = ''.join([x.zfill(2) for x in chirality.split(',')]) \
                + 'r_' + str(zcells) + 'cells' + '.gen'
        elif shape == 'planar' \
                and (xcells != 1 or ycells != 1) and zcells == 1:
            genfile = ''.join([x.zfill(2) for x in chirality.split(',')]) \
                + 'f_' + 'x'.join([str(n) + plural_word_check('cell', n)
                                   for n in (xcells, ycells)]) + '.gen'
        else:
            genfile = ''.join([x.zfill(2) for x in chirality.split(',')]) \
                + ('f' if shape == 'planar' else 'r') + '_' + \
                'x'.join([n + plural_word_check('cell', n)
                          for n in cell_count.split(',')]) + '.gen'

        return genfile

    def _write_genfile(self):
        f = open(self._genfile_name, 'w')
        for k, v in self._kwargs.iteritems():
            f.write('set {!s} {!s}\n'.format(k, v))
        f.write('generate\n')
        f.write('save {!s}{!s}\n'.format(self._genfile_name.split('.')[0],
                                         format_ext[self._fmt]))
        f.close()

    def _tubegen(self):
        retcode = subprocess.call(["tubegen", self._genfile_name])
        if retcode != 0:
            print('tubegen failed')
        else:
            print('tubegen finished with no errors')



# -*- coding: utf-8 -*-
"""
===============================================================================
Fullerene lattice (:mod:`sknano.nanogen._fullerene_bravais_lattice_generators`)
===============================================================================

.. currentmodule:: sknano.nanogen._fullerene_bravais_lattice_generators

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

#import copy
#import itertools
#import sys

#import numpy as np

from ..chemistry import Atoms
#from ..tools.refdata import CCbond

from ._fullerene_generators import FullereneGenerator
#from ._lattice_generators import BravaisLatticeGenerator

__all__ = ['FullereneBravaisLatticeGenerator']


class FullereneBravaisLatticeGenerator(FullereneGenerator):
    u"""Class for generating fullerene bravais lattice structures.

    .. note::

       This class is under development and not functional! Attempting
       to use it raises a `RuntimeError`.

    """

    def __init__(self, n=int, autogen=True, verbose=False):

        raise RuntimeError('This class is not yet implemented.')

        super(FullereneGenerator, self).__init__(with_units=False,
                                                 verbose=verbose)

        if autogen:
            self.generate_structure_data()

    def generate_structure_data(self):
        """Generate structure data."""
        self._structure_atoms = Atoms()

    def save_data(self, fname=None, structure_format=None,
                  rotation_angle=None, rot_axis=None, deg2rad=True,
                  center_CM=True):
        """Save structure data.

        See :py:meth:`~sknano.nanogen.StructureGenerator.save_data` method
        for documentation.

        """
        if fname is None:
            fname = 'C{}'.format(self.Natoms)

        super(FullereneBravaisLatticeGenerator, self).save_data(
            fname=fname, structure_format=structure_format,
            rotation_angle=rotation_angle, rot_axis=rot_axis,
            deg2rad=deg2rad, center_CM=center_CM)



# -*- coding: utf-8 -*-
"""
===============================================================================
Linear algebra functions for transformations (:mod:`sknano.tools._transforms`)
===============================================================================

.. currentmodule:: sknano.tools._transforms

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import numpy as np

__all__ = ['rotation_matrix']


def rotation_matrix(angle, rot_axis='z', deg2rad=False):
    """Generate a rotation matrix.

    Parameters
    ----------
    angle : float
        rotation angle in radians
    rot_axis : {'x', 'y', 'z'}, optional
        rotation axis
    deg2rad : bool, optional
        Angle is in degrees and needs to be converted to radians

    Returns
    -------
    ndarray
        rotation matrix

    """
    if deg2rad:
        angle = np.radians(angle)
    if rot_axis == 'x':
        return np.array([[1, 0, 0],
                         [0, np.cos(angle), -np.sin(angle)],
                         [0, np.sin(angle), np.cos(angle)]])
    elif rot_axis == 'y':
        return np.array([[np.cos(angle), 0, np.sin(angle)],
                         [0, 1, 0],
                         [-np.sin(angle), 0, np.cos(angle)]])
    else:
        return np.array([[np.cos(angle), -np.sin(angle), 0],
                         [np.sin(angle), np.cos(angle), 0],
                         [0, 0, 1]])



# -*- coding: utf-8 -*-
"""
===========================================================================
Helper functions for chirality data (:mod:`sknano.tools._chirality_funcs`)
===========================================================================

.. currentmodule:: sknano.tools._chirality_funcs

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import re

import numpy as np

from ._luts import chiral_type_name_mappings as Ch_types

__all__ = ['cmp_Ch', 'filter_Ch', 'filter_Ch_list', 'filter_key_type_mappings',
           'generate_Ch_list', 'get_Ch_indices', 'get_Ch_type']

filter_key_type_mappings = {}
filter_key_type_mappings['Ch_type'] = str
for k in ('even', 'odd'):
    filter_key_type_mappings[k + '_only'] = bool

for k in ('min_index', 'max_index',
          'min_n', 'max_n',
          'min_m', 'max_m',
          'n', 'm'):
    filter_key_type_mappings[k] = int


def cmp_Ch(Ch1, Ch2):
    """Custom comparator function for sorting chirality lists.

    Parameters
    ----------
    Ch1, Ch2 : {str, tuple}
        2-tuple or 2-tuple string of integers

    Returns
    -------
    int

    """

    if isinstance(Ch1, tuple) and isinstance(Ch2, tuple):
        n1, m1 = Ch1
        Ch1_type = get_Ch_type(Ch1)
        n2, m2 = Ch2
        Ch2_type = get_Ch_type(Ch2)
    elif isinstance(Ch1, str) and isinstance(Ch2, str):
        n1, m1 = get_Ch_indices(Ch1)
        Ch1_type = get_Ch_type(Ch1)
        n2, m2 = get_Ch_indices(Ch2)
        Ch2_type = get_Ch_type(Ch2)

    if Ch1_type == 'AC' and Ch2_type == 'ZZ':
        return 1
    elif Ch1_type == 'ZZ' and Ch2_type == 'AC':
        return -1
    #if dt1 > dt2:
    #    return 1
    #elif dt1 < dt2:
    #    return -1
    else:
        if n1 > n2:
            return 1
        elif n1 < n2:
            return -1
        else:
            if m1 > m2:
                return 1
            elif m1 < m2:
                return -1
            else:
                return 0


def filter_Ch(Ch, even_only=False, odd_only=False, Ch_type=None,
              min_index=None, max_index=None, min_n=None, max_n=None,
              min_m=None, max_m=None, **kwargs):
    """Filter for testing if chirality satisfies given constraint parameters.

    Parameters
    ----------
    even_only, odd_only : bool, optional
    Ch_type : {None, 'armchair', 'zigzag', 'achiral', 'chiral'}, optional
    min_index, max_index : int, optional
    min_n, max_n : int, optional
    min_m, max_m : int, optional

    Returns
    -------
    bool

    """

    if isinstance(Ch, tuple):
        this_Ch_type = get_Ch_type(Ch)
        n, m = Ch
    else:
        n, m = get_Ch_indices(Ch)
        this_Ch_type = get_Ch_type(Ch)

    if even_only:
        if n % 2 == 0 and m % 2 == 0:
            return True
        else:
            return False

    elif odd_only:
        if this_Ch_type != 'ZZ':
            if n % 2 != 0 and m % 2 != 0:
                return True
            else:
                return False
        else:
            if max(n, m) % 2 != 0:
                return True
            else:
                return False

    elif Ch_type is not None:
        if Ch_type in Ch_types:
            Ch_type = Ch_types[Ch_type]

        if this_Ch_type == Ch_type:
            return True
        else:
            return False

    elif min_index is not None and isinstance(min_index, int):
        if this_Ch_type != 'ZZ':
            if n >= min_index and m >= min_index:
                return True
            else:
                return False
        else:
            if max(n, m) >= min_index:
                return True
            else:
                return False

    elif max_index is not None and isinstance(max_index, int):
        if n <= max_index and m <= max_index:
            return True
        else:
            return False

    else:
        return True


def filter_Ch_list(Ch_list, **kwargs):
    """Filter for filtering list of chiralities against set of conditions.

    Parameters
    ----------
    Ch_list : sequence
        list of chiralities
    kwargs : dict, optional
        dictionary of conditions to test each chirality against.

    Returns
    -------
    sequence
        list of chiralities which passed conditions

    """
    return [Ch for Ch in Ch_list if filter_Ch(Ch, **kwargs)]


def get_Ch_indices(Ch):
    """Return the chiral indicies `n` and `m`.

    Parameters
    ----------
    Ch : str

    Returns
    -------
    sequence
        list of chiral indices `n` and `m`

    """
    return [int(c) for c in re.split('[\(\),\s*]+', Ch)[1:-1]]


def get_Ch_type(Ch):
    """Identify the type of nanotube based on its chirality

    Parameters
    ----------
    Ch : {str, tuple}

    Returns
    -------
    str
        `AC` for armchair tubes,
        `ZZ` for zigzag tubes,
        `Ch` for chiral tubes,

    """
    if isinstance(Ch, tuple):
        n, m = Ch
    else:
        n, m = get_Ch_indices(Ch)
    if n == m:
        return 'AC'
    elif n != m and (n == 0 or m == 0):
        return 'ZZ'
    else:
        return 'Ch'


def generate_Ch_list(ns=None, ni=None, nf=None, dn=None,
                     ms=None, mi=None, mf=None, dm=None,
                     chiral_type=None, handedness=None,
                     echo_zsh_str=False):
    """Generate a list of chiralities.

    Parameters
    ----------
    ns, ms : sequence
        list of n and m chiral indices
    ni, nf, dn : int
    mi, mf, dm : int
    chiral_type : {None, 'all', 'achiral', 'chiral', 'armchair', 'zigzag'}
    handedness : {None, 'all', 'left', 'right'}
    echo_zsh_str : bool, optional

    Returns
    -------
    Ch_list : sequence
        list of chiralities

    """
    Ch_list = []
    try:
        if (not ns or (isinstance(ns, list) and ns[0] is None)) \
                and not ni and not nf and \
                (not ms or (isinstance(ns, list) and ms[0] is None)) \
                and not mi and not mf:
            raise TypeError('ns or [ni,] nf[, dn,] and/or '
                            'ms or [mi,] mf[, dm,] must be specified')
        if not ns and isinstance(ms, list) and ms[0] is not None and \
                not ni and not nf and not mi and not mf:
            ns = ms[:]
            ms = None
        elif not ns and (not ms or isinstance(ms, list) and ms[0] is None) \
                and not ni and not nf and \
                (isinstance(mi, int) or isinstance(mf, int)):
            ni = mi
            nf = mf
            mi = mf = None
            if not dn and isinstance(dm, int):
                dn = dm
                dm = None
        if not ns and (isinstance(ni, int) or isinstance(nf, int)):
            try:
                ns = np.arange(ni, nf + 1, dn, dtype=int).tolist()
            except TypeError:
                try:
                    ni, nf, dn = int(float(ni)), int(float(nf)), int(float(dn))
                    ns = np.arange(ni, nf + 1, dn, dtype=int).tolist()
                except (TypeError, ValueError):
                    try:
                        ni, nf = int(float(ni)), int(float(nf))
                        ns = np.arange(ni, nf + 1, dtype=int).tolist()
                    except (TypeError, ValueError):
                        try:
                            nf = int(float(nf))
                            ns = np.arange(nf + 1, dtype=int).tolist()
                        except (TypeError, ValueError):
                            ns = [int(float(ni))]
        if (not ms or isinstance(ms, list) and ms[0] is None) and \
                not mi and not mf:
            if chiral_type in (Ch_types.keys() + Ch_types.values()):
                if chiral_type in ('achiral', 'aCh'):
                    for n in ns:
                        Ch_list.append((n, n))
                    for n in ns:
                        Ch_list.append((n, 0))
                elif chiral_type in ('armchair', 'AC'):
                    for n in ns:
                        Ch_list.append((n, n))
                elif chiral_type in ('zigzag', 'ZZ'):
                    for n in ns:
                        Ch_list.append((n, 0))
                else:
                    for n in ns:
                        m = 1
                        while m < n:
                            if handedness == 'left':
                                Ch_list.append((m, n))
                            elif handedness == 'right':
                                Ch_list.append((n, m))
                            else:
                                Ch_list.append((n, m))
                                Ch_list.append((m, n))
                            m += 1
            else:
                for n in ns:
                    m = 0
                    while m <= n:
                        if (m == 0) or (m == n):
                            Ch_list.append((n, m))
                        else:
                            if handedness == 'left':
                                Ch_list.append((m, n))
                            elif handedness == 'right':
                                Ch_list.append((n, m))
                            else:
                                Ch_list.append((n, m))
                                Ch_list.append((m, n))
                        m += 1
        elif (not ms or isinstance(ms, list) and ms[0] is None) and \
                (isinstance(mi, int) or isinstance(mf, int)):
            try:
                ms = np.arange(mi, mf + 1, dm, dtype=int).tolist()
            except TypeError:
                try:
                    mi, mf, dm = int(float(mi)), int(float(mf)), int(float(dm))
                    ms = np.arange(mi, mf + 1, dm, dtype=int).tolist()
                except (TypeError, ValueError):
                    try:
                        mi, mf = int(float(mi)), int(float(mf))
                        ms = np.arange(mi, mf + 1, dtype=int).tolist()
                    except (TypeError, ValueError):
                        try:
                            mf = int(float(mf))
                            ms = np.arange(mf + 1, dtype=int).tolist()
                        except (TypeError, ValueError):
                            ms = [int(float(mi))]
    except (TypeError, ValueError):
        Ch_list = None
    else:
        if len(Ch_list) == 0 and isinstance(ms, list) and len(ms) != 0:
            for n in ns:
                for m in ms:
                    if (n == 0) and (m == 0):
                        break
                    elif (n == 0) or (m == 0):
                        nm = (n, m)
                        mn = (m, n)
                        if chiral_type not in \
                                ('chiral', 'Ch', 'armchair', 'AC'):
                            if handedness == 'left':
                                if m != 0 and nm not in Ch_list:
                                    Ch_list.append(nm)
                            elif handedness == 'right':
                                if n != 0 and nm not in Ch_list:
                                    Ch_list.append(nm)
                            else:
                                if nm not in Ch_list:
                                    Ch_list.append(nm)
                                if mn not in Ch_list:
                                    Ch_list.append(mn)
                    elif (n == m):
                        Ch = (n, n)
                        if chiral_type not in \
                                ('chiral', 'Ch', 'zigzag', 'ZZ') \
                                and Ch not in Ch_list:
                            Ch_list.append(Ch)
                    else:
                        nm = (n, m)
                        mn = (m, n)
                        if chiral_type not in ('achiral', 'aCh', 'armchair',
                                               'AC', 'zigzag', 'ZZ'):
                            if handedness == 'left':
                                if n < m and nm not in Ch_list:
                                    Ch_list.append(nm)
                            elif handedness == 'right':
                                if n > m and nm not in Ch_list:
                                    Ch_list.append(nm)
                            else:
                                if nm not in Ch_list:
                                    Ch_list.append(nm)
                                if mn not in Ch_list:
                                    Ch_list.append(mn)
    finally:
        if echo_zsh_str and Ch_list is not None:
            print(' '.join([repr(str(Ch)) for Ch in Ch_list]))
        return Ch_list



# -*- coding: utf-8 -*-
"""
===================================================
Core functions (:mod:`sknano.tools._corefuncs`)
===================================================

.. currentmodule:: sknano.tools._corefuncs

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

__all__ = ['check_type']


def check_type(value, allowed_types=()):
    """Check object `value` type against tuple of `allowed_types`.

    Parameters
    ----------
    value : `object`
    allowed_types : tuple
        tuple of allowed classes and/or types

    Raises
    ------
    `TypeError`
        If `value` fails `isinstance` check against `allowed_types`.

    """
    if not isinstance(value, allowed_types):
        raise TypeError('{} does not have an allowed type.\n'.format(value) +
                        '(Allowed Type(s): {})'.format(allowed_types))



# -*- coding: utf-8 -*-
"""
========================================================================
Class for XYZ atom (:mod:`sknano.structure_io.atoms._xyz_atom`)
========================================================================

.. currentmodule:: sknano.structure_io.atoms._xyz_atom

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from ._atom import Atom

__all__ = ['XYZAtom']


class XYZAtom(Atom):
    """`Atom` class for `xyz` format.

    Parameters
    ----------
    element : {str, int}, optional
        A string representation of the element symbol or an integer specifying
        an element atomic number.
    x, y, z : float, optional
        :math:`x, y, z` components of `Atom` position vector relative to
        origin.

    """

    def __init__(self, element=None, x=None, y=None, z=None):

        super(XYZAtom, self).__init__(element=element, x=x, y=y, z=z)



# -*- coding: utf-8 -*-
"""
============================================================================
Base class for structure data atom (:mod:`sknano.structure_io.atoms._atom`)
============================================================================

.. currentmodule:: sknano.structure_io.atoms._atom

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from collections import OrderedDict

import numpy as np

from ...tools import Vector
from ...tools.refdata import atomic_masses, atomic_mass_symbol_map, \
    atomic_numbers, atomic_number_symbol_map, element_symbols

__all__ = ['Atom']


class Atom(object):
    """Base class for structure data atom.

    Parameters
    ----------
    element : {str, int}, optional
        A string representation of the element symbol or an integer specifying
        an element atomic number.
    x, y, z : float, optional
        :math:`x, y, z` coordinates of `Atom`.

    """

    def __init__(self, element=None, mass=None, x=None, y=None, z=None):

        self._symbol = None
        self._Z = None
        self._m = None

        self._r = Vector(x=x, y=y, z=z)

        if isinstance(element, (int, float)):
            self._Z = int(element)
            idx = self._Z - 1
            try:
                self._symbol = element_symbols[idx]
                self._m = atomic_masses[self._symbol]
            except KeyError:
                print('unrecognized element number: {}'.format(element))
        elif isinstance(element, str):
            self._symbol = element
            try:
                self._Z = atomic_numbers[self._symbol]
                self._m = atomic_masses[self._symbol]
            except KeyError:
                print('Unrecognized atomic symbol: {}'.format(element))
        else:
            self._symbol = None
            self._Z = None
            if mass is not None and isinstance(mass, (int, float)):
                try:
                    if isinstance(mass, float):
                        self._symbol = atomic_mass_symbol_map[mass]
                    elif isinstance(mass, int):
                        self._symbol = atomic_number_symbol_map[int(mass / 2)]
                    self._Z = atomic_numbers[self._symbol]
                    self._m = atomic_masses[self._symbol]
                except KeyError:
                    self._symbol = None
                    self._Z = None
                    self._m = mass
            else:
                self._m = 0

        self._attributes = ['symbol', 'Z', 'm', 'r']

        self._atomdict = OrderedDict()
        self._atomdict['symbol'] = self._symbol
        self._atomdict['Z'] = self._Z
        self._atomdict['m'] = self._m
        self._atomdict['x'] = self._r.x
        self._atomdict['y'] = self._r.y
        self._atomdict['z'] = self._r.z

    def __str__(self):
        """Return string representation of atom."""
        atom_str = ''
        for attr in self._attributes:
            atom_str += \
                'Atom {}: {}\n'.format(attr, getattr(self, '_' + attr))
        return atom_str

    @property
    def atomdict(self):
        """Return dictionary of atom attributes."""
        return self._atomdict

    @property
    def Z(self):
        """Atomic number :math:`Z`.

        Returns
        -------
        int
            Atomic number :math:`Z`.
        """
        return self._Z

    @property
    def symbol(self):
        """Element symbol.

        Returns
        -------
        str
            Element symbol.
        """
        return self._symbol

    @property
    def m(self):
        """Atomic mass :math:`m_a` in atomic mass units.

        Returns
        -------
        float
            Atomic mass :math:`m_a` in atomic mass units.
        """
        return self._m

    @property
    def x(self):
        """:math:`x`-coordinate in units of **Angstroms**.

        Returns
        -------
        float
            :math:`x`-coordinate in units of **Angstroms**.

        """
        return self._r.x

    @x.setter
    def x(self, value=float):
        """Set `Atom` :math:`x`-coordinate in units of **Angstroms**.

        Parameters
        ----------
        value : float
            :math:`x`-coordinate in units of **Angstroms**.

        """
        self._r.x = self._atomdict['x'] = value

    @property
    def y(self):
        """:math:`y`-coordinate in units of **Angstroms**.

        Returns
        -------
        float
            :math:`y`-coordinate in units of **Angstroms**.

        """
        return self._r.y

    @y.setter
    def y(self, value=float):
        """Set `Atom` :math:`y`-coordinate in units of **Angstroms**.

        Parameters
        ----------
        value : float
            :math:`y`-coordinate in units of **Angstroms**.

        """
        self._r.y = self._atomdict['y'] = value

    @property
    def z(self):
        """:math:`z`-coordinate in units of **Angstroms**.

        Returns
        -------
        float
            :math:`z`-coordinate in units of **Angstroms**.

        """
        return self._r.z

    @z.setter
    def z(self, value=float):
        """Set `Atom` :math:`z`-coordinate in units of **Angstroms**.

        Parameters
        ----------
        value : float
            :math:`z`-coordinate in units of **Angstroms**.

        """
        self._r.z = self._atomdict['z'] = value

    @property
    def r(self):
        """:math:`x, y, z` coordinates of `Atom` in units of **Angstroms**.

        Returns
        -------
        ndarray
            3-element ndarray of [:math:`x, y, z`] coordinates of `Atom`.

        """
        return self._r.components

    @r.setter
    def r(self, value=np.ndarray):
        """Set :math:`x, y, z` coordinates of `Atom`.

        Parameters
        ----------
        value : array_like
            3-element array of :math:`x, y, z`-coordinates in units of
            **Angstroms**.

        """
        self.x, self.y, self.z = value[0], value[1], value[2]

    def fix_minus_zero_coords(self, epsilon=1.0e-10):
        """Set really really small negative coordinates to zero.

        Set all coordinates with absolute value less than
        epsilon zero so we don't end up with -0.00000
        coordinates in structure data output.

        Parameters
        ----------
        epsilon : float
            smallest allowed absolute value of any :math:`x,y,z` component.

        """
        self._r.fix_minus_zero_components(epsilon=epsilon)

    def rezero_coords(self, epsilon=1.0e-10):
        """Re-zero position coordinates near zero.

        Set position coordinates with absolute value less than `epsilon` to
        zero.

        Parameters
        ----------
        epsilon : float
            smallest allowed absolute value of any :math:`x,y,z` component.

        """
        self._r.rezero_components(epsilon=epsilon)



# -*- coding: utf-8 -*-
"""
====================================================
NanoGen model (:mod:`sknano.nanogen_gui._ng_model`)
====================================================

.. currentmodule:: sknano.nanogen_gui._ng_model

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from ..nanogen import NanotubeBundle, Graphene

__all__ = ['NGModel']


class NGModel(object):
    """:mod:`~sknano.nanogen_gui` MVC model class."""
    def __init__(self):
        self._observers = []
        self._swntbundle = NanotubeBundle(n=10, m=10, bond=1.42)
        self._graphene = Graphene(length=10, width=1, bond=1.42, edge='AC')

    @property
    def n(self):
        """Chiral index :math:`n`"""
        return self._swntbundle.n

    @n.setter
    def n(self, value):
        self._swntbundle.n = value
        self.notify_observers()

    @property
    def m(self):
        """Chiral index :math:`m`"""
        return self._swntbundle.m

    @m.setter
    def m(self, value):
        self._swntbundle.m = value
        self.notify_observers()

    @property
    def nanotube_bond(self):
        return self._swntbundle.bond

    @nanotube_bond.setter
    def nanotube_bond(self, value):
        self._swntbundle.bond = value
        self.notify_observers()

    @property
    def nanotube_element1(self):
        return self._swntbundle.element1

    @nanotube_element1.setter
    def nanotube_element1(self, value):
        self._swntbundle.element1 = value
        self.notify_observers()

    @property
    def nanotube_element2(self):
        return self._swntbundle.element2

    @nanotube_element2.setter
    def nanotube_element2(self, value):
        self._swntbundle.element2 = value
        self.notify_observers()

    @property
    def Lx(self):
        return self._swntbundle.Lx

    @property
    def Ly(self):
        return self._swntbundle.Ly

    @property
    def Lz(self):
        return self._swntbundle.Lz

    @Lz.setter
    def Lz(self, value):
        self._swntbundle.Lz = value
        self.notify_observers()

    @property
    def nx(self):
        return self._swntbundle.nx

    @nx.setter
    def nx(self, value):
        self._swntbundle.nx = value
        self.notify_observers()

    @property
    def ny(self):
        return self._swntbundle.ny

    @ny.setter
    def ny(self, value):
        self._swntbundle.ny = value
        self.notify_observers()

    @property
    def nz(self):
        return self._swntbundle.nz

    @nz.setter
    def nz(self, value):
        self._swntbundle.nz = value
        self.notify_observers()

    @property
    def length(self):
        return self._graphene.length

    @length.setter
    def length(self, value):
        self._graphene.length = value
        self.notify_observers()

    @property
    def width(self):
        return self._graphene.width

    @width.setter
    def width(self, value):
        self._graphene.width = value
        self.notify_observers()

    @property
    def edge(self):
        return self._graphene.edge

    @edge.setter
    def edge(self, value):
        self._graphene.edge = value

    @property
    def nlayers(self):
        return self._graphene.nlayers

    @nlayers.setter
    def nlayers(self, value):
        self._graphene.nlayers = value
        self.notify_observers()

    @property
    def graphene_bond(self):
        return self._graphene.bond

    @graphene_bond.setter
    def graphene_bond(self, value):
        self._graphene.bond = value
        self.notify_observers()

    @property
    def graphene_element1(self):
        return self._graphene.element1

    @graphene_element1.setter
    def graphene_element1(self, value):
        self._graphene.element1 = value
        self.notify_observers()

    @property
    def graphene_element2(self):
        return self._graphene.element2

    @graphene_element2.setter
    def graphene_element2(self, value):
        self._graphene.element2 = value
        self.notify_observers()

    def register_observer(self, observer):
        self._observers.append(observer)

    def remove_observer(self, observer):
        self._observers.remove(observer)

    def notify_observers(self):
        self._swntbundle.compute_bundle_params()
        self._graphene.compute_layer_params()
        for observer in self._observers[:]:
            observer.update_app_view()



# -*- coding: utf-8 -*-
"""
======================================================
JSON format (:mod:`sknano.structure_io._json_format`)
======================================================

.. currentmodule:: sknano.structure_io._json_format

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from ..chemistry import Atom
from ..tools import get_fpath

from ._structure_data import StructureReader, StructureReaderError, \
    StructureWriter, default_comment_line

__all__ = ['JSONDATA', 'JSONReader', 'JSONWriter']


class JSONReader(StructureReader):
    """Class for reading json file format.

    Parameters
    ----------
    fpath : str
        json file

    """
    def __init__(self, fpath=None):
        super(JSONReader, self).__init__(fpath=fpath)

        if fpath is not None:
            self.read()

    def read(self):
        with open(self.fpath, 'r') as f:
            Natoms = int(f.readline().strip())
            self._comment_line = f.readline().strip()
            lines = f.readlines()
            for line in lines:
                s = line.strip().split()
                if len(s) != 0:
                    atom = \
                        Atom(s[0], x=float(s[1]), y=float(s[2]), z=float(s[3]))
                    self._structure_atoms.append(atom)
            if self._structure_atoms.Natoms != Natoms:
                error_msg = '`jsonfile` contained {} atoms '.format(
                    self._structure_atoms.Natoms) + 'but should contain ' + \
                    '{}'.format(Natoms)
                raise StructureReaderError(error_msg)


class JSONWriter(StructureWriter):
    """Class for writing json chemical file format."""

    @classmethod
    def write(cls, fname=None, outpath=None, atoms=None, comment_line=None):
        """Write structure data to file.

        Parameters
        ----------
        fname : str
        outpath : str, optional
            Output path for structure data file.
        atoms : :py:class:`~sknano.chemistry.Atoms`
            An :py:class:`~sknano.chemistry.Atoms` instance.
        comment_line : str, optional
            A string written to the first line of `json` file. If `None`,
            then it is set to the full path of the output `json` file.

        """
        fpath = get_fpath(fname=fname, ext='json', outpath=outpath,
                          overwrite=True, add_fnum=False)
        if comment_line is None:
            comment_line = default_comment_line

        atoms.rezero_coords()

        with open(fpath, 'w') as f:
            for atom in atoms:
                f.write('{:>3s}{:15.8f}{:15.8f}{:15.8f}\n'.format(
                    atom.symbol, atom.x, atom.y, atom.z))


class JSONDATA(JSONReader):
    """Class for reading and writing structure data in JSON data format.

    Parameters
    ----------
    fpath : str, optional

    """
    def __init__(self, fpath=None):
        super(JSONDATA, self).__init__(fpath=fpath)

    def write(self, jsonfile=None):
        """Write json file.

        Parameters
        ----------
        jsonfile : {None, str}, optional

        """
        try:
            if (jsonfile is None or jsonfile == '') and \
                    (self.fpath is None or self.fpath == ''):
                error_msg = '`jsonfile` must be a string at least 1 ' + \
                    'character long.'
                if jsonfile is None:
                    raise TypeError(error_msg)
                else:
                    raise ValueError(error_msg)
            else:
                jsonfile = self.fpath
            JSONWriter.write(fname=jsonfile, atoms=self._structure_atoms,
                             comment_line=self._comment_line)
        except (TypeError, ValueError) as e:
            print(e)



# -*- coding: utf-8 -*-
"""
====================================================
PDB format (:mod:`sknano.structure_io._pdb_format`)
====================================================

.. currentmodule:: sknano.structure_io._pdb_format

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from collections import OrderedDict

#from .atoms import PDBAtom, PDBAtoms
from ..tools import get_fpath

from ._structure_data import StructureReader, StructureReaderError, \
    StructureWriter, StructureFormat, default_comment_line

__all__ = ['PDBData', 'PDBReader', 'PDBWriter', 'PDBFormat']


class PDBReader(StructureReader):
    """Class for reading pdb chemical file format.

    Parameters
    ----------
    fpath : str
        pdb structure file

    """
    def __init__(self, fpath=None):
        super(PDBReader, self).__init__(fpath=fpath)

        if fpath is not None:
            self.read()

    def read(self):
        """Read PDB file."""
        with open(self.fpath, 'r') as f:
            lines = f.readlines()
            for line in lines:
                print(line)


class PDBWriter(StructureWriter):
    """Class for writing pdb chemical file format."""

    @classmethod
    def write(cls, fname=None, outpath=None, atoms=None, comment_line=None):
        """Write structure data to file.

        Parameters
        ----------
        fname : str
        outpath : str, optional
            Output path for structure data file.
        atoms : :py:class:`~sknano.structure_io.atoms.Atoms`
            An :py:class:`~sknano.structure_io.atoms.Atoms` instance.
        comment_line : str, optional

        """
        fpath = get_fpath(fname=fname, ext='pdb', outpath=outpath,
                          overwrite=True, add_fnum=False)
        if comment_line is None:
            comment_line = default_comment_line

        atoms.rezero_coords()

        with open(fpath, 'w') as f:
            for atom in atoms:
                f.write(PDBFormatter.format(atom))


class PDBData(PDBReader):
    """Class for reading and writing structure data in PDB data format.

    Parameters
    ----------
    fpath : str, optional

    """
    def __init__(self, fpath=None):
        try:
            super(PDBData, self).__init__(fpath=fpath)
        except StructureReaderError:
            pass

    def write(self, pdbfile=None):
        pass


class PDBFormat(StructureFormat):
    """Class defining the structure file format for PDB data.

    Parameters
    ----------

    """
    def __init__(self):
        super(PDBFormat, self).__init__()

        self._records = OrderedDict()

        self._records['HEADER'] = header = {}
        header['format'] = ''

        self._records['TITLE'] = title = {}
        title['format'] = ''

        self._records['ATOM'] = atom = {}
        atom['format'] = \
            '{:6}{:>5} {:>4}{:1}{:>3} {:1}{:>4}{:1}' + \
            '{:3}'.format('') + \
            '{:>8.3f}{:>8.3f}{:>8.3f}' + \
            '{:>6.2f}{:>6.2f}' + \
            '{:10}'.format('') + \
            '{:2}{:2}'

        self._records['CONECT'] = {}

        self._records['MASTER'] = {}

        self._records['END'] = None

        self._properties['RECORDS'] = self._records

    @property
    def records(self):
        return self._records


class PDBFormatter(object):
    """Formatter class to convert a `PDBAtom` to a formatted string."""

    @classmethod
    def format(cls):
        pass



# -*- coding: utf-8 -*-
"""
=======================================================================
Helper functions, LUTs, reference data, etc. (:mod:`sknano.tools`)
=======================================================================

.. currentmodule:: sknano.tools

Contents
========

Abstract data structures
-------------------------
.. autosummary::
   :toctree: generated/

   Point
   Vector
   Quaternion

Helper functions for chirality data
------------------------------------
.. autosummary::
   :toctree: generated/

   cmp_Ch
   filter_Ch
   filter_Ch_list
   generate_Ch_list
   get_Ch_indices
   get_Ch_type

'Core' functions that don't fit under any specific category
------------------------------------------------------------
.. autosummary::
   :toctree: generated/

   check_type

Helper functions for I/O
-------------------------
.. autosummary::
   :toctree: generated/

   get_fpath

Helper functions for manipulating strings
------------------------------------------
.. autosummary::
   :toctree: generated/

   plural_word_check

Math functions
---------------
.. autosummary::
   :toctree: generated/

   totient_func

Helper functions for linear algebra transforms
-----------------------------------------------
.. autosummary::
   :toctree: generated/

   rotation_matrix

LUTs (Look-up 'tables' - lists and dictionaries)
-------------------------------------------------
.. autosummary::
   :toctree: generated/

   chiral_type_name_mappings
   xyz_axes

Sub-packages
============
.. autosummary::
   :toctree: generated/

   refdata

"""
from __future__ import division, print_function, absolute_import

__docformat__ = 'restructuredtext'

from ._chirality_funcs import *
from ._corefuncs import *
from ._coremath import *
from ._iofuncs import *
from ._luts import *
from ._mathfuncs import *
from ._strfuncs import *
from ._transforms import *

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
"""
==========================================================================
Graphene structure generators (:mod:`sknano.nanogen._graphene_generators`)
==========================================================================

.. currentmodule:: sknano.nanogen._graphene_generators

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import copy

import numpy as np

from ..chemistry import Atom, Atoms
from ..tools import rotation_matrix
from ..tools.refdata import CCbond

from ._graphene import Graphene
from ._structure_generator import StructureGenerator

__all__ = ['GrapheneGenerator', 'BiLayerGrapheneGenerator']


class GrapheneGenerator(Graphene, StructureGenerator):
    """Class for generating `N`-layer graphene nanostructures.

    Parameters
    ----------
    length : float
        Length of graphene sheet in **nanometers**
    width : float
        Width of graphene sheet in **nanometers**
    edge : {'AC', 'armchair', 'ZZ', 'zigzag'}, optional
        **A**\ rm\ **C**\ hair or **Z**\ ig\ **Z**\ ag edge along
        the `length` of the sheet.
    element1, element2 : {str, int}, optional
        Element symbol or atomic number of basis
        :class:`~sknano.chemistry.Atoms` 1 and 2
    bond : float, optional
        bond length between nearest-neighbor atoms in **Angstroms**.
    nlayers : int, optional
        Number of graphene layers.
    layer_spacing : float, optional
        Distance between layers in **Angstroms**.
    stacking_order : {'AA', 'AB'}, optional
        Stacking order of graphene layers
    autogen : bool, optional
        automatically generate unit cell and full structure
    verbose : bool, optional
        verbose output

    Notes
    -----
    The `GrapheneGenerator` class and its subclasses generate graphene with
    either an armchair or zigzag edge using a 4-atom conventional unit cell.
    If you want to generate graphene as an *unrolled nanotube*, see the
    :class:`~sknano.nanogen.UnrolledNanotubeGenerator` class.

    .. seealso:: :class:`~sknano.nanogen.UnrolledNanotubeGenerator`

    Examples
    --------

    Start an interactive python or ipython session, then import the
    `GrapheneGenerator` class.

    >>> from sknano.nanogen import GrapheneGenerator

    Now generate a **20 nm x 1 nm** armchair edge graphene nano-ribbon.

    >>> ACG = GrapheneGenerator(length=20, width=1, edge='AC')

    Save structure data in `xyz` format:

    >>> ACG.save_data()

    The rendered structure look like:

    .. image:: /images/20nmx1nm_AC_edge.png

    Now let's generate a **20 nm x 1 nm** zigzag edge graphene nano-ribbon.

    >>> ZZG = GrapheneGenerator(length=20, width=1, edge='ZZ')
    >>> ZZG.save_data()

    The rendered structure looks like:

    .. image:: /images/20nmx1nm_ZZ_edge.png

    Now generate **25 nm x 5 nm**, `armchair` edge,
    5 layer, `AB`-stacked graphene.

    >>> ACG_5layers = GrapheneGenerator(length=25, width=5, edge='AC',
    ...                                 nlayers=5)
    >>> ACG_5layers.save_data()

    The rendered structure looks like:

    .. image:: /images/25nmx5nm_5layer_AC_graphene.png

    Now generate single layer, **10 nm x 10 nm** sheet of BN Graphene.

    >>> BN_graphene = GrapheneGenerator(length=10, width=10, edge='AC',
    ...                                 element1='B', element2='N')
    >>> BN_graphene.save_data()

    The rendered structure looks like:

    .. image:: /images/10nmx10nm_single_layer_BN_graphene.png

    Now, just because we can, generate a **5 nm x 5 nm** sheet of
    Uranium-Einsteinium Graphene.

    >>> UEs_graphene = GrapheneGenerator(width=5, length=5, edge='zigzag',
    ...                                  element1='U', element2='Es')
    >>> UEs_graphene.save_data()

    The rendered structure looks like:

    .. image:: /images/5nmx5nm_single_layer_UEs_graphene.png

    """

    def __init__(self, length=None, width=None, edge=None,
                 element1='C', element2='C', bond=CCbond,
                 nlayers=1, layer_spacing=3.35, stacking_order='AB',
                 autogen=True, with_units=False, units=None,
                 verbose=False):

        super(GrapheneGenerator, self).__init__(
            length=length, width=width, edge=edge,
            element1=element1, element2=element2, bond=bond,
            nlayers=nlayers, layer_spacing=layer_spacing,
            stacking_order=stacking_order, with_units=with_units, units=units,
            verbose=verbose)

        if autogen:
            self.generate_unit_cell()
            self.generate_structure_data()

    def generate_unit_cell(self):
        """Generate the graphene unit cell.

        Called automatically if `autogen` is `True`.

        """
        bond = self._bond
        edge = self._edge
        e1 = self._element1
        e2 = self._element2

        self._unit_cell = Atoms()
        # Set up 4 atom basis
        # Leave atom 1 at the origin
        atom1, atom2, atom3, atom4 = Atom(e1), Atom(e2), Atom(e1), Atom(e2)
        if edge == 'AC':
            # Move atom 2 to 2nd basis position
            atom2.x = -np.sqrt(3) / 2 * bond
            atom2.y = bond / 2

            # Move atom 3 from atom 1 along a1 primitive vector
            atom3.x = -np.sqrt(3) / 2 * bond
            atom3.y = 3 / 2 * bond

            # Move atom 4 from atom 2 along a2 primitive vector
            atom4.y = 2 * bond
        else:
            # Move atom 2 to 2nd basis position
            atom2.x = bond

            # Move atom 3 from atom 1 along a1 primitive vector
            atom3.x = 3 / 2 * bond
            atom3.y = np.sqrt(3) / 2 * bond

            # Move atom 4 from atom 2 along a2 primitive vector
            atom4.x = -bond / 2
            atom4.y = np.sqrt(3) / 2 * bond

        for atom in (atom1, atom2, atom3, atom4):
            self._unit_cell.append(atom)
        self._Natoms = self._unit_cell.Natoms

    def generate_structure_data(self):
        """Generate the full structure coordinates."""
        self._structure_atoms = Atoms()
        for nlayer in xrange(self._nlayers):
            layer = Atoms()
            for nx in xrange(self._Nx):
                for ny in xrange(self._Ny):
                    dr = np.array([nx * self._cell.x,
                                   ny * self._cell.y,
                                   nlayer * self._layer_spacing])
                    for atom in self._unit_cell:
                        layer_atom = Atom(atom.symbol)
                        layer_atom.r = atom.r + dr
                        layer.append(layer_atom)

            if self._Natoms_per_layer is None:
                self._Natoms_per_layer = layer.Natoms

            # translate layer to put its center of mass at the origin
            layer.center_CM(r_indices=[0, 1])
            if (nlayer % 2) != 0:
                layer.translate(self._layer_shift.components)

            self._structure_atoms.extend(layer.atoms)

    def save_data(self, fname=None, outpath=None, structure_format=None,
                  rotation_angle=-90, rot_axis='x', deg2rad=True,
                  center_CM=True):
        """Save structure data.

        See :meth:`~sknano.nanogen.StructureGenerator.save_data` method
        for documentation.

        """
        if fname is None:
            dimensions = '{}nmx{}nm'.format(self._length, self._width)
            nlayer = '{}layer'.format(self._nlayers)
            edge = 'AC' if self._edge in ('AC', 'armchair') else 'ZZ'
            atombond = '{}{}'.format(self._element1, self._element2)
            fname_wordlist = (dimensions, nlayer, edge, atombond, 'graphene')
            fname = '_'.join(fname_wordlist)

        #structure_atoms = list(itertools.chain(*self._structure_atoms))
        #structure_atoms = Atoms(structure_atoms)
        if center_CM and self._nlayers > 1:
            self._structure_atoms.center_CM()

        super(GrapheneGenerator, self).save_data(
            fname=fname, outpath=outpath, structure_format=structure_format,
            rotation_angle=rotation_angle, rot_axis=rot_axis,
            deg2rad=deg2rad, center_CM=False)


class BiLayerGrapheneGenerator(GrapheneGenerator):
    """Class for generating bi-layer graphene structures.

    Parameters
    ----------
    length : float
        Length of graphene sheet in **nanometers**
    width : float
        Width of graphene sheet in **nanometers**
    edge : {'AC', 'armchair', 'ZZ', 'zigzag'}, optional
        **A**\ rm\ **C**\ hair or **Z**\ ig\ **Z**\ ag edge along
        the `length` of the sheet.
    element1, element2 : {str, int}, optional
        Element symbol or atomic number of basis
        :class:`~sknano.chemistry.Atoms` 1 and 2
    bond : float, optional
        bond length between nearest-neighbor atoms in **Angstroms**.
    layer_spacing : float, optional
        Distance between layers in **Angstroms**.
    stacking_order : {'AA', 'AB'}, optional
        Stacking order of graphene layers
    rotation_angle : {None, float}, optional
        Rotation angle of second layer specified in degrees.
        If specified in radians, then you must set `deg2rad=False`
    deg2rad : bool, optional
        The rotation angle is specified in degrees and needs to be converted
        to radians.
    autogen : bool, optional
        if `True`, automatically generate unit cell and full structure
    verbose : bool, optional
        verbose output

    Examples
    --------

    Import the BiLayerGrapheneGenerator class

    >>> from sknano.nanogen import BiLayerGrapheneGenerator

    Generate **10 nm** wide by **1 nm** long `AB` stacked
    bilayer-graphene with a `ZZ` edge:

    >>> bi_graphene = BiLayerGrapheneGenerator(length=10, width=1, edge='ZZ')

    Save structure data in `xyz` format:

    >>> bi_graphene.save_data()

    The rendered structure looks like (after rotating 90 degrees so that
    it better fits the page):

    .. image:: /images/10nmx1nm_bilayer.png

    Now generate bilayer-graphene with top layer rotated by 45 degrees.

    >>> rotated_bilayer = BiLayerGrapheneGenerator(length=10, width=10,
    ...                                            edge='armchair',
    ...                                            rotation_angle=45)
    >>> rotated_bilayer.save_data(fname='rotated_bilayer.xyz')

    The rendered structure looks like:

    .. image:: /images/rotated_bilayer.png

    Now generate BN bilayer-graphene with top layer rotated 45 degrees.

    >>> rotated_BN_bilayer = BiLayerGrapheneGenerator(length=10, width=10,
    ...                                               edge='zigzag',
    ...                                               element1='B',
    ...                                               element2='N',
    ...                                               rotation_angle=45)
    >>> rotated_BN_bilayer.save_data(fname='BN_bilayer_rotated_45deg.xyz')

    The rendered structure looks like:

    .. image:: /images/BN_bilayer_rotated_45deg.png

    """

    def __init__(self, length=None, width=None, edge=None,
                 element1='C', element2='C', bond=CCbond,
                 layer_spacing=3.35, stacking_order='AB',
                 rotation_angle=None, deg2rad=True, autogen=True,
                 verbose=False):

        super(BiLayerGrapheneGenerator, self).__init__(
            length=length, width=width, edge=edge,
            element1=element1, element2=element2, bond=bond,
            nlayers=2, layer_spacing=layer_spacing, autogen=False,
            verbose=verbose)

        self._rotation_matrix = None
        if rotation_angle is not None:
            self._rotation_matrix = rotation_matrix(rotation_angle,
                                                    rot_axis='z',
                                                    deg2rad=deg2rad)

        if autogen:
            super(BiLayerGrapheneGenerator, self).generate_unit_cell()
            self.generate_structure_data()

    def generate_structure_data(self):
        """Generate the full structure coordinates."""
        super(BiLayerGrapheneGenerator, self).generate_structure_data()

        if self._rotation_matrix is not None:
            bilayer = copy.deepcopy(self._structure_atoms)
            self._structure_atoms = Atoms()

            z_coords = bilayer.get_coords(as_dict=True)['z']
            z_set = np.asarray(sorted(list(set(z_coords))))
            epsilon = 1e-10

            for n, z in enumerate(z_set):
                layer = Atoms(atoms=bilayer.get_atoms(asarray=True)[
                    np.where(np.abs(z_coords - z) < epsilon)].tolist(),
                    deepcopy=True)
                if (n % 2) != 0:
                    layer.rotate(self._rotation_matrix)

                self._structure_atoms.extend(layer.atoms)



# -*- coding: utf-8 -*-
"""
=============================================
I/O functions (:mod:`sknano.tools._iofuncs`)
=============================================

.. currentmodule:: sknano.tools._iofuncs

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import os
import re
import sys

__all__ = ['get_fpath']


def get_fpath(fname=None, ext=None, outpath=os.getcwd(), overwrite=False,
              add_fnum=True, fnum=None, include_fname=False, fname_only=False,
              verbose=False):
    """Generate modified `fname` string based on chosen parameters.

    Parameters
    ----------
    fname : str
        Name of file, with or without an extension.
    ext : str, optional
        File extension to append to `fname`. If `ext` is None,
        then `fname` is analyzed to see if it likely already has an
        extension. An extension is set to the
        last element in the list of strings returned by
        `fname.split('.')` **if** this list has more than 1 element.
        Otherwise, `ext` will be set to an empty string `''`.
        If `ext` is not None and is a valid string,
        then `fname` is analyzed to see if it already ends with `ext`.
        If `fname.endswith(ext)` is `True` from the start, then `ext` will
        not be duplicately appended.
    outpath : str, optional
        Absolute or relative path for generated output file.
        Default is the absolute path returned by `os.getcwd()`.
    overwrite : bool, optional
        If `True`, overwrite an existing file if it has the same generated
        file path.
    add_fnum : bool, optional
        Append integer number to output file name, starting with **1**.
    fnum : {None, int}, optional
        Starting file number to append if `add_fnum` is `True`.

        .. note::

        If the generated file path exists and `overwrite` is False,
        setting this parameter has no effect.

    include_fname : bool, optional
        If `True`, return `(fpath, fname)` tuple.
    fname_only : bool, optional
        If `True`, return only `fname`.
    verbose : bool, optional
        Show verbose output.

    Returns
    -------
    fpath : str
        The concatenation of `outpath` followed by the updated `fname`.
    (fpath, fname) : tuple (only if `include_fname` is `True`)
        2-tuple of strings `(fpath, fname)`.
    fname : str (only if `fname_only` is `True`)
        Updated `fname`.

    """
    f = None
    if fname is None or fname == '':
        error_msg = '`fname` must be a string at least 1 character long.'
        if fname is None:
            raise TypeError(error_msg)
        else:
            raise ValueError(error_msg)
    else:
        f = fname
        fsplit = f.split('.')
        if ext is None:
            if len(fsplit) > 1:
                ext = '.' + fsplit[-1]
            else:
                ext = ''
        else:
            # check if extension already starts with a '.'
            if not ext.startswith('.'):
                ext = '.' + ext
            # check if file name already ends with extension.
            if f.split('.')[-1] != ext.split('.')[-1]:
                f += ext

    if add_fnum:
        fname = re.split(ext, f)[0]
        if fnum is not None:
            f = fname + '-{:d}'.format(fnum) + ext
        else:
            f = fname + '-1' + ext

    fpath = None

    if outpath is None:
        outpath = os.getcwd()

    try:
        os.makedirs(outpath)
    except OSError:
        if os.path.isdir(outpath):
            pass
        else:
            outpath = os.curdir
    finally:
        fname = f
        fpath = os.path.join(outpath, fname)
        if os.path.isfile(fpath):
            if overwrite:
                try:
                    os.remove(fpath)
                except OSError as e:
                    print(e)
                    sys.exit(1)
                else:
                    if verbose:
                        print(u'overwriting existing file: {}'.format(fname))
            else:
                if add_fnum:
                    while os.path.isfile(fpath):
                        fname = \
                            '-'.join(re.split('-', re.split(ext, f)[0])[:-1])
                        fnum = re.split('-', re.split(ext, f)[0])[-1]
                        f = fname + '-' + str(int(fnum) + 1) + ext
                        fpath = os.path.join(outpath, f)
                    fname = f
                else:
                    print(u'file exists: {}\n'.format(fpath) +
                          u'Set `add_fnum=True` to generate unique `fname`\n'
                          u'or `overwrite=True` to overwrite existing '
                          u'file.')
                    fpath = None

        if verbose:
            print(u'Generated file name: {}'.format(fname))
            print(u'File path: {}'.format(fpath))

        if fname_only:
            return fname
        elif include_fname:
            return fpath, fname
        else:
            return fpath



# -*- coding: utf-8 -*-
"""
====================================================================
LAMMPS data format (:mod:`sknano.structure_io._lammps_data_format`)
====================================================================

.. currentmodule:: sknano.structure_io._lammps_data_format

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from collections import OrderedDict
import os

import numpy as np

#from ..chemistry import Atom
from .atoms import LAMMPSAtom as Atom
from ..tools import get_fpath

from ._structure_data import StructureReader, StructureWriter, \
    StructureConverter, StructureFormat, StructureDataError, \
    default_comment_line

__all__ = ['DATAData', 'DATAReader', 'DATAWriter', 'DATA2XYZConverter',
           'DATAFormat', 'DATAError', 'LAMMPSDATA', 'atom_style_map']


atom_style_map = {}
atom_style_map['angle'] = \
    ['atom-ID', 'molecule-ID', 'atom-type', 'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['atomic'] = \
    ['atom-ID', 'atom-type', 'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['body'] = \
    ['atom-ID', 'atom-type', 'bodyflag', 'mass',
     'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['bond'] = \
    ['atom-ID', 'molecule-ID', 'atom-type', 'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['charge'] = \
    ['atom-ID', 'atom-type', 'q', 'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['dipole'] = \
    ['atom-ID', 'atom-type', 'q', 'x', 'y', 'z',
     'mux', 'muy', 'muz', 'nx', 'ny', 'nz']

atom_style_map['electron'] = \
    ['atom-ID', 'atom-type', 'q', 'spin', 'eradius',
     'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['ellipsoid'] = \
    ['atom-ID', 'atom-type', 'ellipsoidflag', 'density',
     'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['full'] = \
    ['atom-ID', 'molecule-ID', 'atom-type', 'q',
     'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['line'] = \
    ['atom-ID', 'molecule-ID', 'atom-type', 'lineflag', 'density',
     'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['meso'] = \
    ['atom-ID', 'atom-type', 'rho', 'e', 'cv',
     'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['molecular'] = \
    ['atom-ID', 'molecule-ID', 'atom-type', 'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['peri'] = \
    ['atom-ID', 'atom-type', 'volume', 'density',
     'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['sphere'] = \
    ['atom-ID', 'atom-type', 'diameter', 'density',
     'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['template'] = \
    ['atom-ID', 'molecule-ID', 'template-index', 'template-atom',
     'atom-type', 'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['tri'] = \
    ['atom-ID', 'molecule-ID', 'atom-type', 'triangleflag', 'density',
     'x', 'y', 'z', 'nx', 'ny', 'nz']

atom_style_map['wavepacket'] = \
    ['atom-ID', 'atom-type', 'charge', 'spin', 'eradius', 'etag',
     'cs_re', 'cs_im', 'x', 'y', 'z', 'nx', 'ny', 'nz']


class DATAReader(StructureReader):
    """`StructureReader` class for reading `LAMMPS data` file format.

    Parameters
    ----------
    fpath : str
        `LAMMPS data` file path
    atom_style : {'full', 'atomic'}, optional

    """
    def __init__(self, fpath=None, atom_style='full'):
        super(DATAReader, self).__init__(fpath=fpath)

        #self._structure_atoms = LAMMPSAtoms()

        data_format = DATAFormat(atom_style=atom_style)
        self._data_headers = data_format.properties['headers']
        self._data_sections = data_format.properties['sections']
        self._section_properties = data_format.section_properties
        self._section_syntax_dict = data_format.section_syntax_dict

        self._headers = {}
        self._sections = {}
        self._boxbounds = {}

        if fpath is not None:
            self.read()

    @property
    def headers(self):
        """LAMMPS data file headers."""
        return self._headers

    @property
    def sections(self):
        """LAMMPS data file sections."""
        return self._sections

    @property
    def boxbounds(self):
        """Box bounds."""
        return self._boxbounds

    def read(self):
        """Read data file."""
        with open(self.fpath, 'r') as f:
            self._comment_line = f.readline().strip()

            while True:
                line = f.readline().strip()
                if len(line) == 0:
                    continue
                found = False
                for key in self._data_headers.iterkeys():
                    if key in line:
                        found = True
                        self._headers[key] = \
                            [self._data_headers[key]['dtype'](float(s)) for s
                             in [[ss for ss in line.split()][i] for i in
                                 range(self._data_headers[key]['items'])]]
                        if len(self._headers[key]) == 1:
                            # since the list contains only one element,
                            # replace list value with element 0
                            self._headers[key] = self._headers[key][0]
                        break
                if not found:
                    break

            while True:
                found = False
                for section_key, header_key in self._data_sections.iteritems():
                    if section_key in line:
                        found = True
                        f.readline()
                        Nitems = self._headers[header_key]
                        data = []
                        for n in xrange(Nitems):
                            tmp = []
                            line = f.readline().strip().split()
                            for i, props in enumerate(
                                self._section_properties[
                                    section_key].itervalues()):
                                try:
                                    tmp.append(props['dtype'](float(line[i])))
                                except IndexError:
                                    break
                            data.append(tmp)
                        self._sections[section_key] = data[:]
                        #self._sections[section_key] = \
                        #    [[props['dtype'](s) for props in
                        #        self._section_properties[
                        #            section_key].itervalues() for
                        #        s in f.readline().strip().split()] for n in
                        #        xrange(Nitems)]
                        break
                f.readline()
                line = f.readline().strip()
                if len(line) == 0:
                    break

        self._parse_atoms()
        self._parse_atomtypes()
        self._parse_boxbounds()

    def _parse_atoms(self):
        """Populate `Atoms` object with `Atom` objects"""
        atoms_section = self._sections['Atoms']
        atoms_section_syntax = self._section_syntax_dict['Atoms']

        masses_section = self._sections['Masses']
        masses_section_syntax = self._section_syntax_dict['Masses']

        try:
            velocities_section = self._sections['Velocities']
            velocities_section_syntax = self._section_syntax_dict['Velocities']
        except KeyError:
            velocities_section = []
            velocities_section_syntax = {}

        atom_kwargs = {'atomID': None, 'moleculeID': None,
                       'q': None, 'atomtype': None, 'mass': None,
                       'x': None, 'y': None, 'z': None,
                       'vx': None, 'vy': None, 'vz': None}

        for lmps_atom in atoms_section:
            for kw in atom_kwargs.iterkeys():
                if kw in atoms_section_syntax:
                    atom_kwargs[kw] = \
                        lmps_atom[
                            self._section_properties['Atoms'][kw]['index']]
                elif kw in masses_section_syntax:
                    atomtype = \
                        lmps_atom[
                            self._section_properties[
                                'Atoms']['atomtype']['index']]
                    atom_kwargs[kw] = \
                        masses_section[atomtype-1][
                            self._section_properties['Masses'][kw]['index']]
                elif kw in velocities_section_syntax and \
                        len(velocities_section) == len(atoms_section):
                    atomID = \
                        lmps_atom[
                            self._section_properties[
                                'Atoms']['atomID']['index']]
                    for velocity in velocities_section:
                        velocity_atomID = \
                            velocity[self._section_properties[
                                'Velocities']['atomID']['index']]
                        if velocity_atomID == atomID:
                            atom_kwargs[kw] = \
                                velocity[self._section_properties[
                                    'Velocities'][kw]['index']]
                #else:
                #    print('unknown atom keyword: {}'.format(kw))

            atom = Atom(**atom_kwargs)
            self._structure_atoms.append(atom)

    def _parse_atomtypes(self):
        Ntypes = self._structure_atoms.Ntypes
        atomtypes = self._structure_atoms.atomtypes
        if Ntypes != self._headers['atom types']:
            for atomtype in xrange(1, self._headers['atom types'] + 1):
                if atomtype not in atomtypes:
                    mass = self._sections['Masses'][atomtype - 1][
                        self._section_properties['Masses']['mass']['index']]
                    self._structure_atoms.add_atomtype(
                        Atom(atomtype=atomtype, mass=mass))

    def _parse_boxbounds(self):
        for dim in ('x', 'y', 'z'):
            bounds = \
                self._headers[' '.join([dim + lim for lim in ('lo', 'hi')])]
            self._boxbounds[dim] = {'min': bounds[0], 'max': bounds[-1]}

    def get(self, section_key, colnum=None, colname=None, colindex=None):
        """Return section with `section_key`.

        Parameters
        ----------
        section_key : str
        colnum : int, optional
        colname : str, optional
        colindex : int, optional

        Returns
        -------
        sequence

        """
        section_data = None
        try:
            section_data = self._sections[section_key]
            section_syntax = self._section_syntax_dict[section_key]
        except KeyError as e:
            print(e)
        else:
            try:
                colidx = None
                if colnum is not None:
                    colidx = int(colnum - 1)
                elif colname is not None:
                    colidx = \
                        self._section_properties[section_key][colname]['index']
                elif colindex is not None:
                    colidx = int(colindex)
            except (KeyError, TypeError, ValueError) as e:
                print(e)
            else:
                try:
                    colname = section_syntax[colidx]
                    coltype = \
                        self._section_properties[section_key][colname]['dtype']
                    section_data = \
                        np.asarray(
                            section_data, dtype=coltype)[:, colidx].tolist()
                except TypeError:
                    pass
        finally:
            return section_data


class DATAWriter(StructureWriter):
    """`StructureWriter` class for writing `LAMMPS data` file format."""

    @classmethod
    def write(cls, fname=None, outpath=None, fpath=None, atoms=None,
              boxbounds=None, comment_line=None, assume_unique_atoms=False,
              enforce_consecutive_atomIDs=True, pad_box=True, xpad=10.,
              ypad=10., zpad=10., verbose=False):
        """Write structure data to file.

        Parameters
        ----------
        fname : str, optional
            Output file name.
        outpath : str, optional
            Output file path.
        fpath : str, optional
            Full path (directory path + file name) to output data file.
        atoms : :class:`~sknano.chemistry.Atoms`
            An :class:`~sknano.chemistry.Atoms` instance.
        boxbounds : dict, optional
            If `None`, determined automatically from the `atoms` coordinates.
        comment_line : str, optional
            A string written to the first line of `data` file. If `None`,
            then it is set to the full path of the output `data` file.
        assume_unique_atoms : bool, optional
            Check that each :class:`~sknano.chemistry.Atom` in `atoms`
            has a unique :attr:`~sknano.chemistry.Atom.atomID`.
            If the check fails, then assign a unique
            :attr:`~sknano.chemistry.Atom.atomID` to each
            :class:`~sknano.chemistry.Atom`.
            If `assume_unique_atoms` is True, but the atomID's are not unique,
            LAMMPS will not be able to read the data file.
        verbose : bool, optional
            verbose output

        """
        if fpath is None:
            fpath = get_fpath(fname=fname, ext='data', outpath=outpath,
                              overwrite=True, add_fnum=False)
        if comment_line is None:
            comment_line = default_comment_line

        atoms.rezero_coords()

        atomtypes = atoms.atomtypes

        Natoms = atoms.Natoms
        Natoms_width = \
            8 if len(str(Natoms)) <= 12 else len(str(Natoms)) + 4
        Ntypes = atoms.Ntypes
        Ntypes_width = Natoms_width

        atomID_width = len(str(Natoms)) + 1
        atomtype_width = len(str(Ntypes)) + 1

        if (enforce_consecutive_atomIDs and
            atoms.atom_ids.max() != atoms.Natoms) or \
                (not assume_unique_atoms and
                 len(set(atoms.atom_ids)) != atoms.Natoms):
            atoms.assign_unique_ids()

        if boxbounds is None:
            boxbounds = {'x': {'min': None, 'max': None},
                         'y': {'min': None, 'max': None},
                         'z': {'min': None, 'max': None}}

            for i, dim in enumerate(('x', 'y', 'z')):
                boxbounds[dim]['min'] = atoms.coords[:, i].min()
                boxbounds[dim]['max'] = atoms.coords[:, i].max()

        boxpad = {'x': xpad, 'y': ypad, 'z': zpad}
        pad_eps = 0.001
        if pad_box:
            #for dim, pad in boxpad.iteritems():
            for i, dim in enumerate(('x', 'y', 'z')):
                pad = boxpad[dim]
                if abs(boxbounds[dim]['min'] - atoms.coords[:, i].min()) \
                        < pad - pad_eps:
                    boxbounds[dim]['min'] = boxbounds[dim]['min'] - pad
                if abs(boxbounds[dim]['max'] - atoms.coords[:, i].max()) \
                        < pad - pad_eps:
                    boxbounds[dim]['max'] = boxbounds[dim]['max'] + pad

        lohi_width = 0
        for dim in ('x', 'y', 'z'):
            lohi_width = \
                max(lohi_width, len('{:.6f} {:.6f}'.format(
                    boxbounds[dim]['min'], boxbounds[dim]['max'])) + 4)

        with open(fpath, 'w') as f:
            f.write('# {}\n\n'.format(comment_line.lstrip('#').strip()))
            f.write('{}atoms\n'.format(
                '{:d}'.format(Natoms).ljust(Natoms_width)))
            f.write('{}atom types\n\n'.format(
                '{:d}'.format(Ntypes).ljust(Ntypes_width)))

            for dim in ('x', 'y', 'z'):
                f.write('{}{dim}lo {dim}hi\n'.format(
                    '{:.6f} {:.6f}'.format(
                        boxbounds[dim]['min'],
                        boxbounds[dim]['max']).ljust(lohi_width),
                    dim=dim))

            f.write('\nMasses\n\n')
            for atomtype, properties in atomtypes.iteritems():
                f.write('{}{:.4f}\n'.format(
                    '{:d}'.format(atomtype).ljust(Natoms_width),
                    properties['mass']))

            f.write('\nAtoms\n\n')
            for atom in atoms:
                line = ''
                line += "{:>{}}".format(atom.atomID, atomID_width)
                line += "{:>{}}".format(atom.moleculeID, 3)
                line += "{:>{}}".format(atom.atomtype, atomtype_width)
                line += "{:>{}}".format('{:.1f}'.format(atom.q), 4)
                line += "{:>{}}".format('{:f}'.format(atom.x), 14)
                line += "{:>{}}".format('{:f}'.format(atom.y), 14)
                line += "{:>{}}".format('{:f}'.format(atom.z), 14)
                line += "{:>{}}".format('{:d}'.format(atom.nx), 3)
                line += "{:>{}}".format('{:d}'.format(atom.ny), 3)
                line += "{:>{}}".format('{:d}'.format(atom.nz), 3)
                line += '\n'

                f.write(line)

            f.write('\nVelocities\n\n')
            for atom in atoms:
                line = ''
                line += "{:>{}}".format(atom.atomID, atomID_width)
                line += "{:>{}}".format('{:f}'.format(atom.vx), 14)
                line += "{:>{}}".format('{:f}'.format(atom.vy), 14)
                line += "{:>{}}".format('{:f}'.format(atom.vz), 14)
                line += '\n'

                f.write(line)


class DATA2XYZConverter(StructureConverter):
    """
    `StructureConverter` class for converting `LAMMPS data` to `xyz` format.

    .. versionadded:: 0.2.9

    Parameters
    ----------
    datafile : str

    """
    def __init__(self, datafile):
        self._datafile = datafile
        self._xyzfile = os.path.splitext(self._datafile)[0] + '.xyz'

        super(DATA2XYZConverter, self).__init__(
            infile=self._datafile, outfile=self._xyzfile)

    @property
    def datafile(self):
        """`LAMMPS data` file."""
        return self.infile

    @property
    def xyzfile(self):
        """`xyz` file name."""
        return self.outfile

    def convert(self, return_reader=False):
        """Convert `LAMMPS data` to `xyz` chemical file format.

        Parameters
        ----------
        return_reader : bool, optional
            Return an instance of `XYZReader`

        Returns
        -------
        `XYZReader` (only if `return_reader` is True)

        """
        from ._xyz_format import XYZReader, XYZWriter

        datareader = DATAReader(fpath=self.infile)
        atoms = datareader.atoms
        comment_line = datareader.comment_line

        XYZWriter.write(fname=self.outfile, atoms=atoms,
                        comment_line=comment_line)

        if return_reader:
            return XYZReader(fpath=self.outfile)


class DATAData(DATAReader):
    """Class for reading and writing `StructureData` in `LAMMPS data` format.

    Parameters
    ----------
    fpath : str, optional

    """
    def __init__(self, fpath=None):
        super(DATAData, self).__init__(fpath=fpath)

    def delete(self, key):
        pass

    def findtime(self, n):
        pass

    def iterator(self, n):
        pass

    def map(self, *pairs):
        pass

    def maxbox(self):
        pass

    def maxtype(self):
        pass

    def newxyz(self):
        pass

    def reorder(self, colname, *order):
        pass

    def replace(self, section_key, new_data, colnum=None,
                colname=None, colindex=None):
        """Replace section data.

        Parameters
        ----------
        section_key : str
        new_data : sequence
        colnum : int, optional
        colname : str, optional
        colindex : int, optional

        """
        colidx = None

        # for backwards compatibility with the pizza.py data module,
        # first check positional arguments to see if this method was called
        # using the pizza.py data module signature which expects positional
        # arguments of the following type:
        #   data.replace(str, int, list)
        if isinstance(new_data, (int, float)) and \
                isinstance(colnum, (np.ndarray, list)):
            colidx = int(new_data) - 1
            new_data = np.asarray(colnum)
        else:
            try:
                colidx = None
                if colnum is not None:
                    colidx = int(colnum - 1)
                elif colname is not None:
                    colidx = \
                        self._section_properties[section_key][colname]['index']
                elif colindex is not None:
                    colidx = int(colindex)
            except (KeyError, TypeError, ValueError) as e:
                print(e)
                raise DATAError('replace called with invalid arguments')
        atom_attr = self._section_syntax_dict[section_key][colidx]
        attr_dtype = \
            self._section_properties[section_key][atom_attr]['dtype']
        new_data = np.asarray(new_data, dtype=attr_dtype)

        for i, atom in enumerate(self._structure_atoms):
            self._sections[section_key][i][colidx] = \
                attr_dtype(float(new_data[i]))
            setattr(atom, atom_attr, attr_dtype(float(new_data[i])))

    def viz(self, isnap):
        pass

    def write(self, datafile=None):
        """Write data file.

        Parameters
        ----------
        datafile : {None, str}, optional

        """
        try:
            if (datafile is None or datafile == '') and \
                    (self.fpath is None or self.fpath == ''):
                error_msg = \
                    '`datafile` must be a string at least 1 character long.'
                if datafile is None:
                    raise TypeError(error_msg)
                else:
                    raise ValueError(error_msg)
            else:
                datafile = self.fpath

            DATAWriter.write(fname=datafile,
                             atoms=self._structure_atoms,
                             boxbounds=self._boxbounds,
                             comment_line=self._comment_line)
        except (TypeError, ValueError) as e:
            print(e)

    @classmethod
    def format_spec(cls, atom_style='full'):
        return DATAFormat(atom_style=atom_style)

LAMMPSDATA = DATAData


class DATAError(StructureDataError):
    """`StructureDataError` `Exception` for `LAMMPS data` file format.

    Parameters
    ----------
    msg : str
        Error message

    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class DATAFormat(StructureFormat):
    """`StructureFormat` class defining properties for `LAMMPS data` format.

    Parameters
    ----------
    atom_style : {'full'}, optional
        LAMMPS atom style.

    """
    def __init__(self, atom_style='full'):
        super(DATAFormat, self).__init__()
        self._section_properties = OrderedDict()

        atoms_section_syntax = {}
        atoms_section_syntax['angle'] = \
            ['atomID', 'moleculeID', 'atomtype',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['atomic'] = \
            ['atomID', 'atomtype', 'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['body'] = \
            ['atomID', 'atomtype', 'bodyflag', 'mass',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['bond'] = \
            ['atomID', 'moleculeID', 'atomtype',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['charge'] = \
            ['atomID', 'atomtype', 'q',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['dipole'] = \
            ['atomID', 'atomtype', 'q', 'x', 'y', 'z',
             'mux', 'muy', 'muz', 'nx', 'ny', 'nz']

        atoms_section_syntax['electron'] = \
            ['atomID', 'atomtype', 'q', 'spin', 'eradius', 'x', 'y', 'z',
             'nx', 'ny', 'nz']

        atoms_section_syntax['ellipsoid'] = \
            ['atomID', 'atomtype', 'ellipsoidflag', 'density',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['full'] = \
            ['atomID', 'moleculeID', 'atomtype', 'q',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['line'] = \
            ['atomID', 'moleculeID', 'atomtype', 'lineflag', 'density',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['meso'] = \
            ['atomID', 'atomtype', 'rho', 'e', 'cv',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['molecular'] = \
            ['atomID', 'moleculeID', 'atomtype',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['peri'] = \
            ['atomID', 'atomtype', 'volume', 'density',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['sphere'] = \
            ['atomID', 'atomtype', 'diameter', 'density',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['template'] = \
            ['atomID', 'moleculeID', 'template-index', 'template-atom',
             'atomtype', 'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['tri'] = \
            ['atomID', 'moleculeID', 'atomtype', 'triangleflag', 'density',
             'x', 'y', 'z', 'nx', 'ny', 'nz']

        atoms_section_syntax['wavepacket'] = \
            ['atomID', 'atomtype', 'charge', 'spin', 'eradius', 'etag',
             'cs_re', 'cs_im', 'x', 'y', 'z', 'nx', 'ny', 'nz']

        velocities_section_syntax = {}
        velocities_section_syntax['full'] = ['atomID', 'vx', 'vy', 'vz']

        syntax_dtypes = \
            {'atomID': int, 'moleculeID': int, 'atomtype': int, 'q': float,
             'mass': float, 'x': float, 'y': float, 'z': float,
             'nx': int, 'ny': int, 'nz': int,
             'vx': float, 'vy': float, 'vz': float}

        section_keys = ['Atoms', 'Masses', 'Velocities']
        self._section_syntax_dict = \
            {'Atoms': atoms_section_syntax[atom_style],
             'Masses': ['atomtype', 'mass'],
             'Velocities': velocities_section_syntax[atom_style]}

        for section_key in section_keys:
            self._section_properties[section_key] = OrderedDict()
            section_syntax_list = self._section_syntax_dict[section_key]
            for i, syntax in enumerate(section_syntax_list):
                self._section_properties[section_key][syntax] = \
                    {'dtype': syntax_dtypes[syntax],
                     'colnum': i+1,
                     'index': i}

        self._header_keys = ["atoms",
                             "atom types",
                             "bonds",
                             "bond types",
                             "angles",
                             "angle types",
                             "dihedrals",
                             "dihedral types",
                             "impropers",
                             "improper types",
                             "bodies",
                             "ellipsoids",
                             "lines",
                             "triangles",
                             "xlo xhi",
                             "ylo yhi",
                             "zlo zhi",
                             "xy xz yz"]

        self._headers = {'atoms': {'dtype': int, 'items': 1},
                         'atom types': {'dtype': int, 'items': 1},
                         'bonds': {'dtype': int, 'items': 1},
                         'bond types': {'dtype': int, 'items': 1},
                         'angles': {'dtype': int, 'items': 1},
                         'angle types': {'dtype': int, 'items': 1},
                         'dihedrals': {'dtype': int, 'items': 1},
                         'dihedral types': {'dtype': int, 'items': 1},
                         'impropers': {'dtype': int, 'items': 1},
                         'improper types': {'dtype': int, 'items': 1},
                         'bodies': {'dtype': int, 'items': 1},
                         'ellipsoids': {'dtype': int, 'items': 1},
                         'lines': {'dtype': int, 'items': 1},
                         'triangles': {'dtype': int, 'items': 1},
                         'xlo xhi': {'dtype': float, 'items': 2},
                         'ylo yhi': {'dtype': float, 'items': 2},
                         'zlo zhi': {'dtype': float, 'items': 2},
                         'xy xz yz': {'dtype': float, 'items': 3}}

        self._properties['headers'] = self._headers

        self._section_header_map = {}

        atoms_sections = ['Atoms', 'Velocities', 'Molecules']
        self._section_header_map.update(
            dict.fromkeys(atoms_sections, 'atoms'))

        bonds_sections = ['Bonds']
        self._section_header_map.update(
            dict.fromkeys(bonds_sections, 'bonds'))

        lines_sections = ['Lines']
        self._section_header_map.update(
            dict.fromkeys(lines_sections, 'lines'))

        ellipsoids_sections = ['Ellipsoids']
        self._section_header_map.update(
            dict.fromkeys(ellipsoids_sections, 'ellipsoids'))

        triangles_sections = ['Triangles']
        self._section_header_map.update(
            dict.fromkeys(triangles_sections, 'triangles'))

        bodies_sections = ['Bodies']
        self._section_header_map.update(
            dict.fromkeys(bodies_sections, 'bodies'))

        angles_sections = ['Angles']
        self._section_header_map.update(
            dict.fromkeys(angles_sections, 'angles'))

        dihedrals_sections = ['Dihedrals']
        self._section_header_map.update(
            dict.fromkeys(dihedrals_sections, 'dihedrals'))

        impropers_sections = ['Impropers']
        self._section_header_map.update(
            dict.fromkeys(impropers_sections, 'impropers'))

        atom_types_sections = ['Masses', 'Pair Coeffs']
        self._section_header_map.update(
            dict.fromkeys(atom_types_sections, 'atom types'))

        bond_types_sections = ['Bond Coeffs']
        self._section_header_map.update(
            dict.fromkeys(bond_types_sections, 'bond types'))

        angle_types_sections = \
            ['Angle Coeffs', 'BondBond Coeffs', 'BondAngle Coeffs']
        self._section_header_map.update(
            dict.fromkeys(angle_types_sections, 'angle types'))

        improper_types_sections = \
            ['AngleAngle Coeffs', 'Improper Coeffs']
        self._section_header_map.update(
            dict.fromkeys(improper_types_sections, 'improper types'))

        dihedral_types_sections = \
            ['Dihedral Coeffs', 'MiddleBondTorsion Coeffs',
             'EndBondTorsion Coeffs', 'AngleTorsion Coeffs',
             'AngleAngleTorsion Coeffs', 'BondBond13 Coeffs']
        self._section_header_map.update(
            dict.fromkeys(dihedral_types_sections, 'dihedral types'))

        self._properties['sections'] = self._section_header_map

    @property
    def header_keys(self):
        """List of header key names"""
        return self._header_keys

    @property
    def section_properties(self):
        """List of section properties"""
        return self._section_properties

    @property
    def section_syntax_dict(self):
        """Section syntax dictionary."""
        return self._section_syntax_dict



# -*- coding: utf-8 -*-
"""
==============================================================================
Base class for structure data atoms (:mod:`sknano.structure_io.atoms._atoms`)
==============================================================================

.. currentmodule:: sknano.structure_io.atoms._atoms

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import copy
import math

from collections import OrderedDict, MutableSequence

import numpy as np

from ...tools import xyz_axes

__all__ = ['Atoms']


class Atoms(MutableSequence):
    """Base class for collection of `Atom` objects.

    Parameters
    ----------
    atoms : {None, sequence, `Atoms`}, optional
        if not `None`, then a list of `Atom` instance objects or an
        existing `Atoms` instance object.
    copylist : bool, optional
        perform shallow copy of atoms list
    deepcopy : bool, optional
        perform deepcopy of atoms list

    """

    def __init__(self, atoms=None, copylist=True, deepcopy=False,
                 update_property_lists=True):
        self._atoms = []

        self._coords = []
        self._masses = []
        self._symbols = []

        self._property_lists = \
            {'m': self._masses, 'r': self._coords, 'symbol': self._symbols}

        if atoms is not None:
            try:
                if copylist and not deepcopy:
                    self._atoms.extend(atoms.atoms[:])
                elif deepcopy:
                    self._atoms.extend(copy.deepcopy(atoms.atoms))
                else:
                    self._atoms.extend(atoms.atoms)
            except AttributeError:
                if isinstance(atoms, list):
                    if copylist and not deepcopy:
                        self._atoms = atoms[:]
                    elif deepcopy:
                        self._atoms = copy.deepcopy(atoms)
                    else:
                        self._atoms = atoms

                    if update_property_lists:
                        for atom in self._atoms:
                            for p, plist in self._property_lists.iteritems():
                                plist.append(getattr(atom, p))
                else:
                    raise TypeError('`Atoms(atoms={!r})` '.format(atoms) +
                                    'is not a valid `Atoms` constructor '
                                    'argument.\n atoms must be `None`, a list '
                                    'of `Atom` objects, or an `Atoms` object '
                                    'instance.')

    def __str__(self):
        """Return string representation of `Atoms`."""
        atoms_str = ''
        for atom in self._atoms:
            atoms_str += str(atom)
        return atoms_str

    @property
    def atoms(self):
        """Return the list of `Atom` objects"""
        return self._atoms

    @property
    def CM(self):
        """Center-of-Mass coordinates of `Atoms`.

        Returns
        -------
        ndarray
            3-element ndarray specifying center-of-mass coordinates of `Atoms`.

        """
        masses = np.asarray([self.masses])
        coords = np.asarray(self.coords)
        MxR = masses.T * coords
        return np.sum(MxR, axis=0) / np.sum(masses)

    @property
    def M(self):
        """Total mass of `Atoms`."""
        return math.fsum(self.masses)

    @property
    def Natoms(self):
        """Number of atoms in `Atoms`."""
        return len(self._atoms)

    @property
    def coords(self):
        """Return array of `Atom` coordinates."""
        coords = []
        for atom in self._atoms:
            coords.append(atom.r)
        self._coords = coords[:]
        return np.asarray(self._coords)

    @property
    def masses(self):
        """Return the list of `Atom` masses."""
        #self._masses_array = np.asarray(self._masses)
        masses = []
        for atom in self._atoms:
            masses.append(atom.m)
        self._masses = masses[:]
        return self._masses

    @property
    def symbols(self):
        """Return array of `Atom` symbols."""
        symbols = []
        for atom in self._atoms:
            symbols.append(atom.symbol)
        self._symbols = symbols[:]
        return np.asarray(self._symbols)

    def center_CM(self, r_indices=[0, 1, 2]):
        """Center atoms on CM coordinates

        Parameters
        ----------
        r_indices : sequence, optional
            list of component indices of CM position vector to translate
            and center on the origin.

        """
        dr = -self.CM
        self.translate(dr, r_indices=r_indices)

    def clip_bounds(self, min_limit=None, max_limit=None,
                    abs_limit=None, r_indices=[0, 1, 2]):
        """Remove atoms outside the given limits along given dimension.

        Parameters
        ----------
        min_limit : {None, float}, optional
        max_limit : {None, float}, optional
        abs_limit : {None, float}, optional
        r_indices : sequence, optional

        """
        for atom in self._atoms[:]:
            r = atom.r.tolist()
            for i, ri in enumerate(r[:]):
                if i in r_indices:
                    if abs_limit is not None:
                        if abs(ri) <= abs_limit:
                            continue
                        else:
                            del self._atoms[self._atoms.index(atom)]
                            break
                    else:
                        if min_limit is not None:
                            if ri >= min_limit:
                                continue
                            else:
                                del self._atoms[self._atoms.index(atom)]
                                break
                        if max_limit is not None:
                            if ri <= max_limit:
                                continue
                            else:
                                del self._atoms[self._atoms.index(atom)]
                                break

    def fix_minus_zero_coords(self, epsilon=1.0e-10):
        """Set really really small negative coordinates to zero.

        Set all coordinates with absolute value less than
        epsilon zero so we don't end up with -0.00000
        coordinates in structure data output.

        Parameters
        ----------
        epsilon : float
            smallest allowed absolute value of any component of an
            atom's position

        """
        for atom in self._atoms:
            r = atom.r.tolist()
            for i, ri in enumerate(r[:]):
                if ri < 0 and abs(ri) < epsilon:
                    r[i] = 0.0
            atom.x, atom.y, atom.z = r

    def getatomsattr(self, asarray=False, as_array=False):
        pass

    def get_atoms(self, asarray=False, as_array=False):
        """Return list of `Atoms`.

        Parameters
        ----------
        asarray, as_array : bool, optional

        Returns
        -------
        sequence or ndarray

        """
        if asarray or as_array:
            return np.asarray(self._atoms)
        else:
            return self._atoms

    def get_coords(self, components=None, as_dict=False):
        """Return atom coords.

        Parameters
        ----------
        components : {None, sequence}, optional
        as_dict : bool, optional

        Returns
        -------
        coords : :py:class:`python:~collections.OrderedDict` or ndarray

        """
        coords = self.coords
        if as_dict:
            if components is None or components == 'r':
                components = ('x', 'y', 'z')
            elif isinstance(components, str):
                components = (components,)

            return OrderedDict(zip(
                components, [coords[:, xyz_axes.index(component)] for
                             component in components]))
        else:
            return coords

    def rezero_coords(self, epsilon=1.0e-10):
        """Set really really small coordinates to zero.

        Set all coordinates with absolute value less than
        epsilon to zero.

        Parameters
        ----------
        epsilon : float
            smallest allowed absolute value of any :math:`x,y,z` component.

        """
        for atom in self._atoms:
            r = atom.r.tolist()
            for i, ri in enumerate(r[:]):
                if abs(ri) < epsilon:
                    r[i] = 0.0
            atom.x, atom.y, atom.z = r

    def rezero_xyz(self, epsilon=1.0e-10):
        return self.rezero_coords(epsilon=epsilon)

    def rotate(self, R_matrix):
        """Rotate atom coordinates using rotation matrix `R_matrix`.

        Parameters
        ----------
        R_matrix : array_like
            3x3 array representation of *3D rotation matrix*.

        """
        for atom in self._atoms:
            atom.r = np.dot(R_matrix, atom.r.T).T

    def translate(self, dr, r_indices=[0, 1, 2]):
        """Translate atom coordinates.

        Parameters
        ----------
        dr : array_like
            array representation of displacement vector to translate
            components of atoms position vector through
        r_indices : sequence, optional
            list of component indices of position vector that
            `dr` vector acts upon

        """
        for atom in self._atoms:
            r = atom.r.tolist()
            for i, ri in enumerate(r[:]):
                if i in r_indices:
                    r[i] += dr[i]
            atom.x, atom.y, atom.z = r

    def __delitem__(self, index):
        """Concrete implementation of @abstractmethod.

        Delete list element `self.atoms[index]` and delete all elements
        from atom properties lists `self.masses[index]`,
        `self.charges[index]`, and `self.coords[index]`

        Parameters
        ----------
        index : int
            index of target list element

        """
        del self._atoms[index]
        for plist in self._property_lists.itervalues():
            del plist[index]

    def __getitem__(self, index):
        """Concrete implementation of @abstractmethod.

        Get `Atom` object instance at list element `self.atoms[index]`

        Parameters
        ----------
        index : int
            index of target list element

        Returns
        -------
        `Atom`
            `Atom` instance at target `self.atoms[index]`

        """
        return self._atoms[index]

    def __len__(self):
        """Concrete implementation of @abstractmethod.

        Returns
        -------
        int
            length of `self.atoms` list.

        """
        return len(self._atoms)

    def __setitem__(self, index, atom):
        """Concrete implementation of @abstractmethod.

        set target list element `self.atoms[index] = atom`

        Also set element of all atom properties lists (`self.masses[index]`,
        `self.charges[index]`, and `self.coords[index]`) to atom instance
        properties (`atom.m`, `atom.q`, `atom.r`), respectively.

        Parameters
        ----------
        index : int
            index of target list element
        atom : `Atom`
            `Atom` instance object to set target list element to.

        """
        self._atoms[index] = atom
        for p, plist in self._property_lists.iteritems():
            plist[index] = getattr(atom, p)

    def insert(self, index, atom):
        """Concrete implementation of @abstractmethod.

        insert `Atom` instance at target list `index`

        Also insert `Atom` instance properties at the given target list index
        for all `Atom` properties in `self._property_lists.keys()`
        into their respective target lists of `Atoms` properties
        `self._property_lists.values()`.

        Parameters
        ----------
        index : int
            index of target list element
        atom : `Atom`
            `Atom` object instance to set target list element to

        """
        self._atoms.insert(index, atom)
        for p, plist in self._property_lists.iteritems():
            plist.insert(index, getattr(atom, p))
